from __future__ import annotations

import os
from pathlib import Path
from subprocess import Popen

from bmde.core import logging
from bmde.core.exec import ExecOptions

from ...config.loader import load_settings
from ...core.logging import configure_logging_from_settings
from ...core.spec_opts import SpecExecOpts
from ...core.types import BackendOptions
from .service import GitService
from .settings import GitSettings
from .spec import GitSpec, GitSpecOpts

log = logging.get_logger(__name__)


def create_git_spec(
    d: Path | None,
    arguments: list[str] | None = None,
    backend: BackendOptions | None = None,
    interactive: bool = True,
    dry_run: bool = True,
    entrypoint: Path | None = None,
    ssh_username: str | None = None,
    ssh_password: str | None = None,
    ssh_host: str | None = None,
    git_name: str | None = None,
    git_email: str | None = None,
    vpn_username: str | None = None,
    vpn_password: str | None = None,
    vpn_host: str | None = None,
    vpn_port: int | None = None,
    vpn_realm: str | None = None,
    vpn_cert: str | None = None,
    vpn_test_dns: str | None = None,
    vpn_test_ip: str | None = None,
    settings: GitSettings | None = None,
) -> GitSpec:
    if d is None:
        d = Path(os.getcwd())

    if settings is None:
        settings = GitSettings()

    return GitSpec(
        GitSpecOpts=GitSpecOpts(
            d=d,
            ssh_username=(
                ssh_username if ssh_username is not None else settings.ssh.username
            ),
            ssh_password=(
                ssh_password if ssh_password is not None else settings.ssh.password
            ),
            ssh_host=ssh_host if ssh_host is not None else settings.ssh.host,
            git_name=git_name if git_name is not None else settings.git.name,
            git_email=git_email if git_email is not None else settings.git.email,
            vpn_username=(
                vpn_username if vpn_username is not None else settings.vpn.username
            ),
            vpn_password=(
                vpn_password if vpn_password is not None else settings.vpn.password
            ),
            vpn_host=vpn_host if vpn_host is not None else settings.vpn.host,
            vpn_port=vpn_port if vpn_port is not None else settings.vpn.port,
            vpn_realm=vpn_realm if vpn_realm is not None else settings.vpn.realm,
            vpn_cert=vpn_cert if vpn_cert is not None else settings.vpn.cert,
            vpn_test_dns=(
                vpn_test_dns if vpn_test_dns is not None else settings.vpn.test_dns
            ),
            vpn_test_ip=(
                vpn_test_ip if vpn_test_ip is not None else settings.vpn.test_ip
            ),
        ),
        SpecExecOpts=SpecExecOpts(
            backend=(
                backend
                if backend is not None
                else (
                    settings.execution_settings.backend
                    if settings.execution_settings.backend is not None
                    else BackendOptions.DOCKER
                )
            ),
            entrypoint=(
                entrypoint
                if entrypoint is not None
                else settings.execution_settings.entrypoint
            ),
            arguments=arguments,
            dry_run=(
                dry_run if dry_run is not None else settings.execution_settings.dry_run
            ),
            background=(
                settings.execution_settings.background
                if settings.execution_settings.background is not None
                else False
            ),
            interactive=(
                interactive
                if interactive is not None
                else settings.execution_settings.interactive
            ),
        ),
    )


def execute_git(spec: GitSpec) -> int | Popen[bytes]:
    return GitService().run(
        spec.GitSpecOpts,
        ExecOptions(
            dry_run=spec.SpecExecOpts.dry_run,
            background=spec.SpecExecOpts.background,
            interactive=spec.SpecExecOpts.interactive,
            entrypoint=spec.SpecExecOpts.entrypoint,
            arguments=spec.SpecExecOpts.arguments,
            backend=spec.SpecExecOpts.backend,
        ),
    )


def git_command(
    d: Path | None,
    arguments: list[str] | None = None,
    backend: BackendOptions | None = None,
    dry_run: bool = False,
    interactive: bool = True,
    entrypoint: Path | None = None,
    ssh_username: str | None = None,
    ssh_password: str | None = None,
    ssh_host: str | None = None,
    git_name: str | None = None,
    git_email: str | None = None,
    vpn_username: str | None = None,
    vpn_password: str | None = None,
    vpn_host: str | None = None,
    vpn_port: int | None = None,
    vpn_realm: str | None = None,
    vpn_cert: str | None = None,
    vpn_test_dns: str | None = None,
    vpn_test_ip: str | None = None,
    settings: GitSettings | None = None,
) -> int | Popen[bytes]:

    if settings is None:
        full_settings = load_settings()
        configure_logging_from_settings(settings=full_settings)
        settings = full_settings.git

    spec = create_git_spec(
        d=d,
        arguments=arguments,
        backend=backend,
        dry_run=dry_run,
        interactive=interactive,
        entrypoint=entrypoint,
        ssh_username=ssh_username,
        ssh_password=ssh_password,
        ssh_host=ssh_host,
        git_name=git_name,
        git_email=git_email,
        vpn_username=vpn_username,
        vpn_password=vpn_password,
        vpn_host=vpn_host,
        vpn_port=vpn_port,
        vpn_realm=vpn_realm,
        vpn_cert=vpn_cert,
        vpn_test_dns=vpn_test_dns,
        vpn_test_ip=vpn_test_ip,
        settings=settings,
    )

    return execute_git(spec)



import os
import subprocess

from bmde.core import logging
from bmde.core.docker import can_run_docker
from bmde.core.exec import ExecOptions, run_cmd
from bmde.core.os_utils import host_uid_gid

from ..spec import BuildSpecOpts
from .backend import BuildBackend

log = logging.get_logger(__name__)


class DockerRunner(BuildBackend):
    def is_available(self) -> bool:
        return can_run_docker()

    def run(self, spec: BuildSpecOpts, exec_opts: ExecOptions) -> int | subprocess.Popen[bytes]:

        args = ["docker",
                    "run",
                    "--pull=always",  # Always update image to last version
                    "--rm"]  # Remove container after execution (one-off container)

        if exec_opts.interactive:
            args += ["-it"]  # Interactive mode with the container CLI

        if exec_opts.entrypoint:
            args += ["--entrypoint", str(exec_opts.entrypoint)]

        uid, gid = host_uid_gid()
        args += ["--user", f"{uid}:{gid}"]  # Make the generated files owned by the same user running the container

        project_path = os.path.basename(spec.d)
        args += ["-v", f"{spec.d}:/input/{project_path}:rw"]
        args += ["-w", f"/input/{project_path}"]  # Start with working directory in project mounted

        args += ["aleixmt/bmde-linux:latest"]  # Base image

        if exec_opts.arguments is not None:
            args += list(exec_opts.arguments)

        return run_cmd(args, exec_opts)



from dataclasses import dataclass
from pathlib import Path

from bmde.commands.run.spec import RunSpec
from bmde.core.spec import BaseSpec
from bmde.core.spec_opts import SpecExecOpts
from bmde.core.types import DockerOutputOptions


@dataclass
class DebugSpecOpts(BaseSpec):
    elf: Path
    docker_screen: DockerOutputOptions | None
    docker_network: str | None
    RunSpec: RunSpec


@dataclass
class DebugSpec(BaseSpec):
    SpecExecOpts: SpecExecOpts
    DebugSpecOpts: DebugSpecOpts



import typer

from bmde.commands.git.command import git_command
from bmde.config.schema import Settings
from bmde.core import logging
from bmde.core.shared_options import (
    ArgumentsOpt,
    BackendOpt,
    DirectoryOpt,
    DryRunOpt,
    EntrypointOpt,
    GitEmailOpt,
    GitNameOpt,
    InteractiveOpt,
    SshHostOpt,
    SshPasswordOpt,
    SshUsernameOpt,
    VpnCertOpt,
    VpnHostOpt,
    VpnPasswordOpt,
    VpnPortOpt,
    VpnRealmOpt,
    VpnTestDnsOpt,
    VpnTestIpOpt,
    VpnUsernameOpt,
)

log = logging.get_logger(__name__)


def git_controller(
    ctx: typer.Context,
    arguments: ArgumentsOpt = None,
    directory: DirectoryOpt = None,
    backend: BackendOpt = None,
    entrypoint: EntrypointOpt = None,
    dry_run: DryRunOpt = False,
    interactive: InteractiveOpt = True,
    ssh_username: SshUsernameOpt = None,
    ssh_password: SshPasswordOpt = None,
    ssh_host: SshHostOpt = None,
    git_name: GitNameOpt = None,
    git_email: GitEmailOpt = None,
    vpn_username: VpnUsernameOpt = None,
    vpn_password: VpnPasswordOpt = None,
    vpn_host: VpnHostOpt = None,
    vpn_port: VpnPortOpt = None,
    vpn_realm: VpnRealmOpt = None,
    vpn_cert: VpnCertOpt = None,
    vpn_test_dns: VpnTestDnsOpt = None,
    vpn_test_ip: VpnTestIpOpt = None,
) -> None:
    """git wrapper with SSH password bypass and VPN management. git is a distributed version control system."""

    log.trace(
        "CLI options provided:\n"
        f"- Arguments: {str(arguments)}\n"
        f"- Directory: {str(directory)}\n"
        f"- Backend: {str(backend)}\n"
        f"- Entrypoint: {str(entrypoint)}\n"
        f"- Dry run: {str(dry_run)}\n"
        f"- Interactive: {str(interactive)}\n"
        f"- SSH username: {ssh_username}\n"
        f"- SSH password: {ssh_password}\n"
        f"- SSH host: {ssh_host}\n"
        f"- git name: {git_name}\n"
        f"- git email: {git_email}\n"
        f"- VPN username: {vpn_username}\n"
        f"- VPN password: {vpn_password}\n"
        f"- VPN host: {vpn_host}\n"
        f"- VPN port: {str(vpn_port)}\n"
        f"- VPN realm: {vpn_realm}\n"
        f"- VPN cert: {vpn_cert}\n"
        f"- VPN test DNS: {vpn_test_dns}\n"
        f"- VPN test IP: {vpn_test_ip}\n"
    )

    settings: Settings = ctx.obj["settings"]

    log.trace("Final settings for git command:\n" f"- {str(settings.git)}\n")

    git_command(
        d=directory,
        arguments=arguments,
        backend=backend,
        dry_run=dry_run,
        interactive=interactive,
        entrypoint=entrypoint,
        ssh_username=ssh_username,
        ssh_password=ssh_password,
        ssh_host=ssh_host,
        git_name=git_name,
        git_email=git_email,
        vpn_username=vpn_username,
        vpn_password=vpn_password,
        vpn_host=vpn_host,
        vpn_port=vpn_port,
        vpn_realm=vpn_realm,
        vpn_cert=vpn_cert,
        vpn_test_dns=vpn_test_dns,
        vpn_test_ip=vpn_test_ip,
        settings=settings.git,
    )



from pathlib import Path

from pydantic import BaseModel

from bmde.config.command_settings import ExecutionSettings
from bmde.core.types import DOCKER_DESMUME_DEBUG_NETWORK, DockerOutputOptions


class RunSettings(BaseModel):
    graphical_output: DockerOutputOptions | None = DockerOutputOptions.HOST

    execution_settings: ExecutionSettings = ExecutionSettings()

    debug: bool | None = False
    arm9_debug_port: int | None = 1024
    fat_image: Path | None = None
    docker_network: str | None = DOCKER_DESMUME_DEBUG_NETWORK



import typer

from bmde.commands.debug.command import debug_command
from bmde.config.schema import Settings
from bmde.core import logging
from bmde.core.shared_options import (
    ArgumentsOpt,
    BackendOpt,
    BackgroundOpt,
    DockerScreenOpt,
    DryRunOpt,
    ElfRomOpt,
    EntrypointOpt,
    NdsRomOpt,
    PortOpt,
)

log = logging.get_logger(__name__)


def debug_controller(
    ctx: typer.Context,
    nds: NdsRomOpt = None,
    elf: ElfRomOpt = None,
    arguments: ArgumentsOpt = None,
    docker_screen: DockerScreenOpt = None,
    # common flags
    backend: BackendOpt = None,
    background: BackgroundOpt = False,
    entrypoint: EntrypointOpt = None,
    port: PortOpt = 1024,
    dry_run: DryRunOpt = False,
) -> None:
    """desmume wrapper. Runs an NDS ROM."""

    settings: Settings = ctx.obj["settings"]

    log.debug(
        "CLI options provided:\n"
        f"- Arguments: {str(arguments)}\n"
        f"- Backend: {str(backend)}\n"
        f"- Background: {str(background)}\n"
        f"- Entrypoint: {str(entrypoint)}\n"
        f"- Docker screen: {str(docker_screen)}\n"
        f"- NDS ROM: {str(nds)}\n"
    )

    log.debug(
        "Final settings for debug command:\n"
        f"- Arguments: {str(arguments)}\n"
        f"- Backend: {str(backend if backend is not None else settings.debug.execution_settings.backend)}\n"
        f"- Background: {str(background)}\n"
        f"- Entrypoint: {str(entrypoint if entrypoint is not None else settings.debug.execution_settings.entrypoint)}\n"
        f"- Dry run: {str(dry_run)}\n"
        f"- Docker screen: {str(docker_screen if docker_screen is not None else settings.debug.docker_screen)}\n"
        f"- NDS ROM: {str(nds)}\n"
    )

    ret = debug_command(
        nds=nds,
        elf=elf,
        arguments=arguments,
        backend=backend,
        background=background,
        docker_screen=docker_screen,
        dry_run=dry_run,
        entrypoint=entrypoint,
        port=port,
        settings=settings.debug,
    )

    if isinstance(ret, int):
        raise typer.Exit(ret)



import shutil
import subprocess

from bmde.core.exec import ExecOptions, run_cmd

from ..spec import RunSpecOpts
from .backend import RunBackend


class FlatpakRunner(RunBackend):
    def is_available(self) -> bool:
        return shutil.which("flatpak") is not None

    def run(
        self, spec: RunSpecOpts, exec_opts: ExecOptions
    ) -> int | subprocess.Popen[bytes]:
        # Adjust to your flatpak package and args
        args = ["flatpak", "run", "org.desmume.DesmuME", str(spec.nds_rom)]
        return run_cmd(args, exec_opts)



from collections.abc import Iterable
from pathlib import Path


def find_upwards(name: str, start: Path) -> Path | None:
    """
    Finds a file with the given name in the start path and in all of its predecessors and returns its path. If not,
    returns None.

    Args:
        name: Name of the desired file.
        start: Directory where the function start looking for the file.

    Returns:
        The path file or None if no such file exists.
    """
    cur = start.resolve()
    root = cur.anchor
    while True:
        candidate = cur / name
        if candidate.exists():
            return candidate
        if str(cur) == root:
            return None
        cur = cur.parent


def find_first(paths: Iterable[Path], pattern: str) -> Path | None:
    for p in paths:
        for hit in p.glob(pattern):
            return hit
    return None



#!/usr/bin/env python3
import re
from pathlib import Path

INCLUDE_RE = re.compile(
    r'^\s*--8<--\s+"([^":]+)(?::(\d+)?:(\d+)?)?"\s*$'
)

def expand_include(line: str, base_dir: Path) -> list[str]:
    """
    Expands a MkDocs-style include directive into file contents.
    Supports optional :start:end line ranges.
    """
    match = INCLUDE_RE.match(line)
    if not match:
        return [line]

    path, start, end = match.groups()
    #print(f"start {str(start)} end {str(end)} path {str(path)}")
    file_path = (base_dir / path).resolve()

    if not file_path.exists():
        raise FileNotFoundError(f"Included file not found: {file_path}")

    with open(file_path, encoding="utf-8") as f:
        lines = f.readlines()

    # Apply line slicing if specified (1-based, inclusive)
    if start or end:
        s = int(start) - 1 if start else 0
        e = int(end) if end else len(lines)
        lines = lines[s:e]

    lines.append("\n")
    return lines


def render_file(src: Path, dst: Path):
    base_dir = src.parent.parent.parent
    output = []

    with open(src, encoding="utf-8") as f:
        for line in f.readlines():
            expanded = expand_include(line, base_dir)
            output.extend(expanded)

    dst.write_text("".join(output))


def main():
    SRC = "docs/getting-started/host_components_installation.md"
    DST = "docs/getting-started/host_components_installation_substituted.md"

    src = Path(SRC)
    dst = Path(DST)

    render_file(src, dst)
    print(f"Generated {dst} from {src}")


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from pathlib import Path

from bmde.core.spec import BaseSpec
from bmde.core.spec_opts import SpecExecOpts
from bmde.core.types import DockerOutputOptions


@dataclass
class RunSpecOpts(BaseSpec):
    nds_rom: Path
    arm9_debug_port: int | None
    debug: bool
    docker_network: str
    fat_image: Path | None
    graphical_output: DockerOutputOptions | None


@dataclass
class RunSpec(BaseSpec):
    SpecExecOpts: SpecExecOpts
    RunSpecOpts: RunSpecOpts



import logging
from logging import Logger
from pathlib import Path
from typing import Any, cast

from rich.logging import RichHandler

from bmde.config.schema import Settings
from bmde.core.types import DATE_FORMAT, LogLevel, get_default_log_path

# ---- extend the logging module with TRACE
TRACE_LEVEL_NUM = 1
logging.addLevelName(TRACE_LEVEL_NUM, "TRACE")
MESSAGE_FORMAT = "%(asctime)s | %(name)s | %(message)s"


class ExtendedLogger(logging.Logger):
    def trace(self: Logger, message: str, *args: Any, **kwargs: Any) -> None:
        self._log(TRACE_LEVEL_NUM, message, args, **kwargs)


# Tell the logging system to use your new class
logging.setLoggerClass(ExtendedLogger)


class SecretsFilter(logging.Filter):
    def __init__(self, secrets: list[str | None] | None):
        super().__init__()
        self.secrets: list[str | None] = secrets or []

    def filter(self, record: logging.LogRecord) -> bool:
        if not self.secrets:
            return True

        if isinstance(record.msg, str):
            for secret in self.secrets:
                if secret and secret in record.msg:
                    record.msg = record.msg.replace(secret, "*****")

        """
        if record.args:
            if isinstance(record.args, tuple):
                new_args = []
                modified = False
                for arg in record.args:
                    temp_arg = arg
                    for secret in self.secrets:
                        if secret and secret in temp_arg:
                            temp_arg = temp_arg.replace(secret, "*****")
                            modified = True
                    new_args.append(temp_arg)
                if modified:
                    record.args = tuple(new_args)
        """
        return True


def setup_logging(
    level: int | None | None,
    log_file: str | Path | None = None,
    secrets: list[str | None] | None = None,
) -> None:
    # Default level is INFO
    if level is None:
        level = logging.INFO

    handlers: list[logging.Handler] = []
    secrets_filter = SecretsFilter(secrets)

    # ---- console (Rich)
    console = RichHandler(
        rich_tracebacks=True,
        markup=True,
        show_time=False,
        show_level=True,
        show_path=False,
    )
    common_formatter = logging.Formatter(MESSAGE_FORMAT, DATE_FORMAT)
    console.setLevel(level)
    console.setFormatter(common_formatter)
    console.addFilter(secrets_filter)
    handlers.append(console)

    if log_file:
        path = Path(log_file)
        path.parent.mkdir(parents=True, exist_ok=True)
        file_handler = logging.FileHandler(path, encoding="utf-8")
        file_handler.setLevel(level)
        file_handler.setFormatter(common_formatter)
        file_handler.addFilter(secrets_filter)
        handlers.append(file_handler)

    logging.basicConfig(
        level=level,  # root captures everything
        handlers=handlers,
        format=MESSAGE_FORMAT,
        force=True,
    )


def obfuscate_text(text: str | None) -> str:
    if text is None:
        return str(text)
    else:
        return "*****"


def get_logger(name: str) -> ExtendedLogger:
    """Return a logger with trace() method available."""
    return cast(ExtendedLogger, logging.getLogger(name))


def process_log_flags(
    very_verbose: bool, verbose: bool, quiet: bool, very_quiet: bool
) -> tuple[LogLevel | None, bool]:
    more_than_one_flag = False
    flag_counter = 0
    for flag in (very_verbose, verbose, quiet, very_quiet):
        if flag:
            flag_counter += 1
    if flag_counter > 1:
        more_than_one_flag = True

    if very_verbose:
        return LogLevel.TRACE, more_than_one_flag
    elif verbose:
        return LogLevel.DEBUG, more_than_one_flag
    elif quiet:
        return LogLevel.WARNING, more_than_one_flag
    elif very_quiet:
        return LogLevel.QUIET, more_than_one_flag
    else:
        return None, more_than_one_flag


def configure_logging_from_settings(settings: Settings) -> None:
    if settings.logging.file is None:
        settings.logging.file = get_default_log_path()

    if settings.logging.level is None:
        settings.logging.level = LogLevel.get_default_log_level()

    # Always obfuscate all passwords
    words_to_obfuscate = [settings.git.vpn.password, settings.git.ssh.password]
    # Optionally but by default true obfuscate other sensible data.
    if settings.logging.hide_sensibles:
        words_to_obfuscate += [
            settings.git.git.name,
            settings.git.git.email,
            settings.git.ssh.host,
            settings.git.ssh.username,
            settings.git.vpn.host,
            str(settings.git.vpn.port),
            settings.git.vpn.username,
            settings.git.vpn.cert,
            settings.git.vpn.realm,
            settings.git.vpn.test_dns,
            settings.git.vpn.test_ip,
        ]

    setup_logging(
        level=settings.logging.level.to_logging_level(),
        log_file=settings.logging.file,
        secrets=words_to_obfuscate,
    )  # Preventive creation of log for logging the loading of settings



from __future__ import annotations

import subprocess
from abc import ABC
from subprocess import Popen
from typing import Any, Generic, TypeVar

from ..core import logging
from ..core.exec import ExecOptions
from .backend import Backend
from .spec import BaseSpec
from .types import BackendOptions

log = logging.get_logger(__name__)

SpecType = TypeVar("SpecType", bound=BaseSpec)  # will be GitSpec, BuildSpec, etc.
BackendType = TypeVar(
    "BackendType", bound=Backend[Any], covariant=True
)  # will be BuildBackend, RunBackend, etc.


class Service(Generic[SpecType, BackendType], ABC):

    def __init__(
        self,
        order: list[BackendOptions | BackendOptions],
        mapping: dict[BackendOptions | BackendOptions, BackendType],
    ) -> None:
        # We should use Any for a good generic class, but I prefer to break inheritance and restrict certain types
        # because I can not use a parent, since Enums are final (RunBackendOptions and BackendOptions
        self.order = order
        self.mapping = mapping

    def choose_backend(
        self, env: BackendOptions | BackendOptions | None
    ) -> list[BackendType]:
        order = self.order
        if env:
            order = [env]  # force
        return [self.mapping[e] for e in order]

    def run(self, spec: SpecType, exec_opts: ExecOptions) -> int | Popen[bytes]:
        for backend in self.choose_backend(exec_opts.backend):
            if backend.is_available():
                ret = backend.run(spec, exec_opts)
                # Use isinstance for type narrowing
                if isinstance(ret, (int, subprocess.Popen)):
                    return ret
                    # returncode is guaranteed to be an int after communicate()
        raise RuntimeError("No suitable backend available")



"""
Helpers to parse strings & map log levels.
"""

from __future__ import annotations

import datetime
import logging
import os
import pathlib
from enum import Enum

log = logging.getLogger(
    __name__
)  # Types cannot use logging module to avoid circular imports

PROJECT_DIR = pathlib.Path(__file__).resolve().parent.parent.parent.parent
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
NOW = datetime.datetime.now().strftime(DATE_FORMAT)
# Only used by run and debug modules in Docker mode
DOCKER_DESMUME_DEBUG_NETWORK = "bmde-debug"


class BackendOptions(Enum):
    """
    Common environment backends.

    Note:
    - "bmde" is included for future/host-like flows (your Bare Metal Dev Env).
    """

    HOST = "host"
    DOCKER = "docker"
    FLATPAK = "flatpak"

    @classmethod
    def parse(cls, value: str | None) -> BackendOptions | None:
        """Parse case-insensitively; returns None if value is falsy."""
        log.debug("Executing function parse from Backend")
        if not value:
            return None
        norm = value.strip().lower()
        try:
            return cls(norm)  # Enum accepts the value directly
        except ValueError as exc:
            valid = ", ".join(v.value for v in cls)
            raise ValueError(f"Unknown environment '{value}'. Valid: {valid}") from exc


class DockerOutputOptions(Enum):
    """
    Execution environment backends. Obtained by composition with Backend class
    """

    VNC = "vnc"
    HOST = "host"
    NONE = "none"

    @classmethod
    def parse(cls, value: str | None) -> DockerOutputOptions | None:
        """Parse case-insensitively; returns None if value is falsy."""
        log.debug("Executing function parse from DockerOutputName")
        if not value:
            return None
        norm = value.strip().lower()
        try:
            return cls(norm)  # Enum accepts the value directly
        except ValueError as exc:
            valid = ", ".join(v.value for v in cls)
            raise ValueError(f"Unknown environment '{value}'. Valid: {valid}") from exc


class LogLevel(Enum):
    """
    Logical log levels for the CLI.

    Includes a custom TRACE (more verbose than DEBUG) and QUIET
    (suppresses all output beyond CRITICAL).
    """

    TRACE = "trace"
    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    QUIET = "quiet"

    @classmethod
    def parse(cls, value: str | None) -> LogLevel | None:
        """Parse case-insensitively; returns None if value is falsy."""
        print("Executing function parse from LogLevel")
        if not value:
            return None
        norm = value.strip().lower()
        try:
            return cls(norm)
        except ValueError as exc:
            valid = ", ".join(v.value for v in cls)
            raise ValueError(f"Unknown log level '{value}'. Valid: {valid}") from exc

    @classmethod
    def get_default_log_level(cls) -> LogLevel:
        return LogLevel.INFO

    def to_logging_level(self) -> int:
        if self is LogLevel.TRACE:
            return 0
        if self is LogLevel.DEBUG:
            return logging.DEBUG
        if self is LogLevel.INFO:
            return logging.INFO
        if self is LogLevel.WARNING:
            return logging.WARNING
        if self is LogLevel.ERROR:
            return logging.ERROR
        if self is LogLevel.QUIET:
            return logging.CRITICAL + 10
        # Fallback
        return logging.INFO


def get_default_log_path() -> pathlib.Path:
    return pathlib.Path(str(os.path.join(".", "logs", "bmde", NOW + ".log")))



from __future__ import annotations

import subprocess
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from bmde.core.exec import ExecOptions
from bmde.core.spec import BaseSpec

SpecType = TypeVar("SpecType", bound=BaseSpec)  # will be GitSpec, BuildSpec, etc.


class Backend(ABC, Generic[SpecType]):
    """
    Generic interface for backends. Each runner backend must implement a function to use it and to determine if it
    is available.
    """

    @abstractmethod
    def is_available(self) -> bool: ...

    def check(self) -> bool:
        """
        Performs detailed checks to verify if the backend is correctly configured.
        Returns True if all checks pass, False otherwise.
        Default implementation returns True.
        """
        return True

    @abstractmethod
    def run(
        self, spec: SpecType, exec_opts: ExecOptions
    ) -> int | subprocess.Popen[bytes]: ...



from __future__ import annotations

import os
import tomllib
from pathlib import Path
from typing import Any

from ..core import logging
from ..core.paths import find_upwards
from .schema import Settings

log = logging.get_logger(__name__)


def read_toml(path: Path) -> dict[str, str]:
    """
    Reads .toml file with its paths and returns a dictionary with its values.

    Args:
        path: Path of the TOML file.

    Returns: Plain dictionary with the key-values of the TOML file.

    """
    if not path or not path.is_file():
        log.debug(f"File {path} not found or not a valid file")
        return {}
    with path.open("rb") as f:
        log.debug(f"Reading TOML file: {path}")
        settings = tomllib.load(f)
        log.trace(f"Loaded settings: {settings}")
        return settings


def env_config(prefix: str = "BMDE_") -> dict[str, dict[str, str]]:
    """
    Reads the configuration from the environment and returns it as a nested dictionary of the read subcommands.

    Args:
        prefix: The prefix that the system variables must have to be added to the dictionary.

    Returns:
        Nested dictionary. The first key is the word after the first _ of the name of the system variable in lowercase,
        which represents the subcommand name (run, build...) and the second key is the rest of the name of the system
        variable in lowercase. The value is the value of the property. (e.g., BMDE_RUN_ENVIRONMENT=docker -->
        env[RUN][ENVIRONMENT] = docker)

    """
    result: dict[str, dict[str, str]] = {}
    for k, v in os.environ.items():
        if not k.startswith(prefix):
            continue
        parts = k[len(prefix) :].lower().split("_")
        if len(parts) >= 2:
            section, key = parts[0], "_".join(parts[1:])
            if section not in result:
                result[section] = {}
            result[section][key] = v
    log.debug(f"Parsed environment variables: {str(result)}")
    return result


def merge(a: dict[Any, Any], b: dict[Any, Any]) -> dict[Any, Any]:
    """
    Merges two dictionaries by setting the values of the first dictionary a with the values of the second dictionary b
    for all the coinciding keys.
    If the values for the same key in the two dictionaries are also dictionaries, perform recursive call to merge those
    in the same way.

    Args:
        a: The first dictionary to merge.
        b: The second dictionary to merge.

    Returns:
        The merged dictionary.
    """
    out = dict(a)
    for k, v in b.items():
        if isinstance(v, dict) and isinstance(out.get(k), dict):
            out[k] = merge(out[k], v)
        else:
            out[k] = v
    return out


def load_settings(explicit_config: Path | None = None) -> Settings:
    """
    Loads settings following the priority:
    1. Environment variables: Variables prefixed with BMDE_ENVIRONMENT
    2. /etc/bmde/bmde.toml
    3. ~/.config/bmde/bmde.toml
    4. bmde.toml from the repository
    5. Explicit config given via arguments

    This behaviour is NOT performed by this function but for completion we must say that settings with priority 6 would
    come from CLI arguments.

    Args:
        explicit_config: Path to a run-specific BMDE's TOML configuration file.

    Returns:
        Settings object with the configuration from all subcommands.

    """
    # 1) system env (lowest)
    log.debug("Trying to read global config from Environment Variables")
    acc = env_config()

    global_paths = [
        Path("/etc/bmde/bmde.toml"),  # 2) system-wide global config
        Path(
            os.path.expanduser("~/.config/bmde/bmde.toml")
        ),  # 3) user-specific global config
    ]
    for gp in global_paths:
        log.debug(f"Trying to read global config from {gp}")
        acc = merge(acc, read_toml(gp))

    # 4) repo config (closest bmde.toml up-tree)
    start_path = Path.cwd()
    log.debug(
        f'Trying to find repo config "bmde.toml" starting from {start_path} and upwards.'
    )
    repo_cfg = find_upwards("bmde.toml", start=start_path)
    if repo_cfg:
        log.debug(f"Trying to read repo config found in {repo_cfg}")
        acc = merge(acc, read_toml(repo_cfg))

    # 5) execution-specific config (highest)
    if explicit_config:
        log.debug(f"Trying to read explicit CLI config from {explicit_config}")
        acc = merge(acc, read_toml(explicit_config))

    return Settings.model_validate(acc)



from dataclasses import dataclass
from pathlib import Path

from bmde.core.spec import BaseSpec
from bmde.core.types import BackendOptions


@dataclass
class SpecExecOpts(BaseSpec):
    backend: BackendOptions
    background: bool
    dry_run: bool
    interactive: bool
    entrypoint: Path | None
    arguments: list[str] | None



import shutil
import subprocess

from bmde.core import logging
from bmde.core.exec import ExecOptions, run_cmd
from bmde.core.os_utils import is_command_available

from ..spec import RunSpecOpts
from .backend import RunBackend

log = logging.get_logger(__name__)


class HostRunner(RunBackend):
    def is_available(self) -> bool:
        return self.check()

    def check(self) -> bool:
        available = is_command_available("desmume") or is_command_available(
            "desmume-cli"
        )
        if not available:
            log.error("desmume or desmume-cli not found in PATH.")
        return available

    def run(
        self, spec: RunSpecOpts, exec_opts: ExecOptions
    ) -> int | subprocess.Popen[bytes]:
        if exec_opts.entrypoint is not None:
            entry = str(exec_opts.entrypoint)
        else:
            desmume_path = shutil.which("desmume")
            desmumecli_path = shutil.which("desmume-cli")
            if desmume_path is not None:
                entry = desmume_path
            elif desmumecli_path is not None:
                entry = desmumecli_path
            else:
                entry = "desmume"
        args = [entry, str(spec.nds_rom)]
        if spec.fat_image:
            args += ["--cflash-image", str(spec.fat_image)]
        if spec.debug:
            args += ["--arm9gdb-port", str(spec.arm9_debug_port), "--gdb-stub"]
        if exec_opts.arguments is not None:
            args += list(exec_opts.arguments)
        return run_cmd(args, exec_opts)



from pathlib import Path

from pydantic import BaseModel

from bmde.core.types import BackendOptions


class ExecutionSettings(BaseModel):
    entrypoint: Path | None = None
    arguments: list[str] | None = None
    background: bool | None = False
    dry_run: bool | None = False
    interactive: bool = True
    backend: BackendOptions | None = None



import os
import subprocess

from bmde.core import logging
from bmde.core.docker import can_run_docker
from bmde.core.exec import ExecOptions, run_cmd

from ..spec import PatchSpecOpts
from .backend import PatchBackend

log = logging.get_logger(__name__)


class DockerRunner(PatchBackend):
    def is_available(self) -> bool:
        return can_run_docker()

    def run(
        self, spec: PatchSpecOpts, exec_opts: ExecOptions
    ) -> int | subprocess.Popen[bytes]:
        entry = []
        if exec_opts.entrypoint:
            entry = ["--entrypoint", str(exec_opts.entrypoint)]

        docker_img = "aleixmt/dlditool:latest"

        mounts = [
            "-v",
            f"{spec.nds_rom.parent}:/input/{os.path.basename(spec.nds_rom.parent)}:rw",
        ]
        workdir_opt = [
            "-w",
            f"/input/{os.path.basename(spec.nds_rom.parent)}",
        ]  # Workdir

        arguments = []
        if exec_opts.arguments is not None:
            arguments = list(exec_opts.arguments)
        else:
            arguments = [
                "mpcf.dldi",
                f"/input/{os.path.basename(spec.nds_rom.parent)}/{os.path.basename(spec.nds_rom)}",
            ]

        run_args = ["docker", "run", "--pull=always", "--rm"]

        if exec_opts.interactive:
            run_args += ["-it"]

        run_args += [
            *mounts,
            *entry,
            *workdir_opt,
            docker_img,
            *arguments,
        ]

        log.debug("Patch arguments for Docker backend" + str(run_args))
        return run_cmd(run_args, exec_opts)



from __future__ import annotations

import subprocess
from dataclasses import dataclass
from pathlib import Path

from bmde.core import logging
from bmde.core.types import BackendOptions

log = logging.get_logger(__name__)


@dataclass
class ExecOptions:
    dry_run: bool | None | None = False
    env: dict[str, str] | None = None
    cwd: str | None = None
    background: bool | None = False
    interactive: bool = True
    backend: BackendOptions | None = None
    entrypoint: Path | None = None
    arguments: list[str] | None = None


def run_cmd(cmd: list[str], opts: ExecOptions) -> int | subprocess.Popen[bytes]:
    if isinstance(cmd, str):
        pretty = cmd
        args = cmd
    else:
        pretty = " ".join(cmd)
        args = cmd

    log.trace("exec: %s", pretty)
    if opts.dry_run:
        log.info("[dry-run] %s", pretty)
        return 0

    if opts.background:
        # Non-blocking execution
        proc = subprocess.Popen(
            args,
            env=opts.env,
            cwd=opts.cwd,
        )
        return proc  # caller can manage it
    else:
        return subprocess.call(
            args, env=opts.env, cwd=opts.cwd
        )  # no shell injection risk



from pydantic import BaseModel

from bmde.commands.run.settings import RunSettings
from bmde.config.command_settings import ExecutionSettings
from bmde.core.types import DockerOutputOptions


class DebugSettings(BaseModel):
    run: RunSettings = RunSettings()
    execution_settings: ExecutionSettings = ExecutionSettings()

    docker_screen: DockerOutputOptions | None = DockerOutputOptions.HOST
    docker_network: str | None = None



from __future__ import annotations

from bmde.commands.run.backends.backend import RunBackend
from bmde.core import logging
from bmde.core.service import Service
from bmde.core.types import BackendOptions as RunBackendName

from .backends.docker import DockerRunner
from .backends.flatpak import FlatpakRunner
from .backends.host import HostRunner
from .spec import RunSpecOpts

log = logging.get_logger(__name__)


class RunService(Service[RunSpecOpts, RunBackend]):
    def __init__(self) -> None:
        super().__init__(
            [RunBackendName.DOCKER, RunBackendName.FLATPAK, RunBackendName.HOST],
            {
                RunBackendName.HOST: HostRunner(),
                RunBackendName.DOCKER: DockerRunner(),
                RunBackendName.FLATPAK: FlatpakRunner(),
            },
        )



import typer

from bmde.config.loader import load_settings
from bmde.core import logging
from bmde.core.logging import (
    configure_logging_from_settings,
    process_log_flags,
)
from bmde.core.shared_options import (
    ConfigOpt,
    LogFileOpt,
    QuietOpt,
    ShowConfigOpt,
    ShowDefaultConfigOpt,
    VerboseOpt,
    VeryQuietOpt,
    VeryVerboseOpt,
)

log = logging.get_logger(__name__)


def cli_global_callback(
    ctx: typer.Context,
    config: ConfigOpt = None,
    verbose: VerboseOpt = False,
    very_verbose: VeryVerboseOpt = False,
    quiet: QuietOpt = False,
    very_quiet: VeryQuietOpt = False,
    log_file: LogFileOpt = None,
    show_config: ShowConfigOpt = False,
    show_default_config: ShowDefaultConfigOpt = False,
) -> None:
    """
    Global option callback. Executed if no command is provided.
    """

    cli_log_level, more_than_one_flag = process_log_flags(
        very_verbose=very_verbose, verbose=verbose, quiet=quiet, very_quiet=very_quiet
    )

    if more_than_one_flag:
        log.warning(
            "More than one log level arguments was provided, the log level with more verbosity will be used."
        )
    # Load settings
    settings = load_settings(explicit_config=config)

    # CLI overrides
    if cli_log_level:
        settings.logging.level = cli_log_level
    if log_file:
        settings.logging.file = log_file

    configure_logging_from_settings(settings=settings)

    # Load settings into global context
    ctx.obj = {"settings": settings}
    log.debug("Ended global callback")



import typer

from bmde.commands.check.command import check_command
from bmde.config.schema import Settings


def check_controller(ctx: typer.Context) -> None:
    """Check availability of backends for all services."""
    settings: Settings = ctx.obj["settings"]
    check_command(settings)



from __future__ import annotations

import os
from pathlib import Path
from subprocess import Popen

from bmde.core import logging
from bmde.core.exec import ExecOptions

from ...config.loader import load_settings
from ...core.logging import configure_logging_from_settings
from ...core.spec_opts import SpecExecOpts
from ...core.types import BackendOptions
from .service import BuildService
from .settings import BuildSettings
from .spec import BuildSpec, BuildSpecOpts

log = logging.get_logger(__name__)


def create_build_spec(
    d: Path | None,
    arguments: list[str] | None = None,
    backend: BackendOptions | None = None,
    background: bool = False,
    interactive: bool = True,
    dry_run: bool = False,
    entrypoint: Path | None = None,
    settings: BuildSettings | None = None,
) -> BuildSpec:
    """
    Creates the BuildSpec from the provided arguments.
    """
    if d is None:
        d = Path(os.getcwd())

    if settings is None:
        settings = BuildSettings()

    # Default args are prepended to CLI args
    passed_args: list[str] = []
    if arguments is not None:
        passed_args += arguments
    if settings.execution_settings.arguments is not None:
        passed_args += settings.execution_settings.arguments

    return BuildSpec(
        BuildSpecOpts=BuildSpecOpts(
            d=d,
        ),
        SpecExecOpts=SpecExecOpts(
            backend=backend if backend is not None else (settings.execution_settings.backend
                                                         if settings.execution_settings.backend is not None else
                                                         BackendOptions.DOCKER),
            entrypoint=(
                entrypoint
                if entrypoint is not None
                else settings.execution_settings.entrypoint
            ),
            arguments=passed_args,
            dry_run=(
                dry_run if dry_run is not None else settings.execution_settings.dry_run
            ),
            background=(
                background
                if background is not None
                else settings.execution_settings.background
            ),
            interactive=(
                interactive
                if interactive is not None
                else settings.execution_settings.interactive
            ),
        ),
    )


def execute_build(spec: BuildSpec) -> int | Popen[bytes]:
    """
    Executes the BuildSpec.
    """
    return BuildService().run(
        spec.BuildSpecOpts,
        ExecOptions(
            arguments=spec.SpecExecOpts.arguments,
            dry_run=spec.SpecExecOpts.dry_run,
            background=spec.SpecExecOpts.background,
            interactive=spec.SpecExecOpts.interactive,
            backend=spec.SpecExecOpts.backend,
            entrypoint=spec.SpecExecOpts.entrypoint
        ),
    )


def build_command(
    d: Path | None,
    arguments: list[str] | None = None,
    backend: BackendOptions | None = None,
    background: bool = False,
    interactive: bool = True,
    dry_run: bool = False,
    entrypoint: Path | None = None,
    settings: BuildSettings | None = None,
) -> int | Popen[bytes]:
    """
    CLI Entrypoint. Orchestrates spec creation and execution.
    """
    if settings is None:
        full_settings = load_settings()
        configure_logging_from_settings(
            settings=full_settings
        )
        settings = full_settings.build

    spec = create_build_spec(
        d=d,
        arguments=arguments,
        backend=backend,
        background=background,
        interactive=interactive,
        dry_run=dry_run,
        entrypoint=entrypoint,
        settings=settings,
    )

    return execute_build(spec)



from bmde.commands.build.command import build_command  # noqa: F401
from bmde.commands.git.command import git_command  # noqa: F401
from bmde.commands.patch.command import patch_command  # noqa: F401
from bmde.commands.run.command import run_command  # noqa: F401



"""
The responsibility of this file is to implement the functions that will be called when calling each subcommand in the
CLI. Each function is responsible for attending a single subcommand with their arguments specified through the function
typehint arguments using Typer. Each function will build a Settings object with the CLI overrides, arguments and default
settings, which will be passed to the function and run a command.
Each function will also be responsible for aborting execution with CLI arguments that are impossible. This only applies
to data coming from the CLI, the syntax of the overrides is not responsibility of the function of these files.
"""

from __future__ import annotations

import typer
from rich.console import Console

import bmde.commands.build.cli
import bmde.commands.check.cli
import bmde.commands.debug.cli
import bmde.commands.git.cli
import bmde.commands.patch.cli
import bmde.commands.run.cli

# Import modules directly to ensure they are loaded and available
import bmde.core.cli_global

console = Console()
app = typer.Typer(
    add_completion=False, help="BMDE CLI", no_args_is_help=True
)  # TODO Completion does not work

# Use the imported modules directly
app.callback()(bmde.core.cli_global.cli_global_callback)

app.command("build")(bmde.commands.build.cli.build_controller)
app.command("git")(bmde.commands.git.cli.git_controller)
app.command("patch")(bmde.commands.patch.cli.patch_controller)
app.command("run")(bmde.commands.run.cli.run_controller)
app.command("debug")(bmde.commands.debug.cli.debug_controller)
app.command("check")(bmde.commands.check.cli.check_controller)

if __name__ == "__main__":
    app()



"""
The responsibility of this file and the similar ones under each command, is to translate the logic of the bmde interface
into the specification of the pure logic of the command, for example, translating shell into a corresponding entrypoint
"""

from __future__ import annotations

from pathlib import Path
from subprocess import Popen

from bmde.core import logging
from bmde.core.exec import ExecOptions

from ...config.loader import load_settings
from ...core.file_utils import resolve_elf, resolve_nds
from ...core.logging import configure_logging_from_settings
from ...core.spec_opts import SpecExecOpts
from ...core.types import (
    DOCKER_DESMUME_DEBUG_NETWORK,
    BackendOptions,
    DockerOutputOptions,
)
from ..run.service import RunService
from ..run.spec import RunSpec, RunSpecOpts
from .service import DebugService
from .settings import DebugSettings
from .spec import DebugSpec, DebugSpecOpts

log = logging.get_logger(__name__)


def create_debug_spec(
    nds: Path | None,
    elf: Path | None,
    arguments: list[str] | None = None,
    backend: BackendOptions | None = None,
    background: bool = False,
    docker_screen: DockerOutputOptions | None = None,
    dry_run: bool = False,
    entrypoint: Path | None = None,
    port: int = 1024,
    settings: DebugSettings | None = None,
) -> DebugSpec:
    if settings is None:
        settings = DebugSettings()

    elf_resolved, _ = resolve_elf(elf, cwd=Path.cwd())
    nds_resolved, _ = resolve_nds(nds, cwd=Path.cwd())

    # Default args are prepended to CLI args
    passed_args: list[str] = []
    if arguments is not None:
        passed_args += arguments
    if settings.execution_settings.arguments is not None:
        passed_args += settings.execution_settings.arguments

    return DebugSpec(
        DebugSpecOpts=DebugSpecOpts(
            elf=elf_resolved,
            docker_screen=(
                docker_screen if docker_screen is not None else settings.docker_screen
            ),
            docker_network=(
                settings.docker_network
                if settings.docker_network
                else DOCKER_DESMUME_DEBUG_NETWORK
            ),
            RunSpec=RunSpec(
                RunSpecOpts=RunSpecOpts(
                    nds_rom=nds_resolved,
                    fat_image=settings.run.fat_image,
                    graphical_output=settings.run.graphical_output,
                    debug=True,
                    arm9_debug_port=(
                        port
                        if port is not None
                        else (
                            settings.run.arm9_debug_port
                            if settings.run.arm9_debug_port is not None
                            else 1024
                        )
                    ),
                    docker_network=(
                        settings.run.docker_network
                        if settings.run.docker_network
                        else DOCKER_DESMUME_DEBUG_NETWORK
                    ),
                ),
                SpecExecOpts=SpecExecOpts(
                    dry_run=dry_run,
                    backend=(
                        settings.run.execution_settings.backend
                        if settings.run.execution_settings.backend is not None
                        else BackendOptions.DOCKER
                    ),
                    background=True,
                    entrypoint=settings.run.execution_settings.entrypoint,
                    arguments=arguments,
                    interactive=False,
                ),
            ),
        ),
        SpecExecOpts=SpecExecOpts(
            backend=(
                backend
                if backend is not None
                else (
                    settings.execution_settings.backend
                    if settings.execution_settings.backend is not None
                    else BackendOptions.DOCKER
                )
            ),
            entrypoint=(
                entrypoint
                if entrypoint is not None
                else settings.execution_settings.entrypoint
            ),
            arguments=passed_args,
            dry_run=(
                dry_run if dry_run is not None else settings.execution_settings.dry_run
            ),
            background=(
                background
                if background is not None
                else settings.execution_settings.background
            ),
            interactive=False,
        ),
    )


def execute_debug(spec: DebugSpec) -> int | Popen[bytes]:
    handle = RunService().run(
        spec.DebugSpecOpts.RunSpec.RunSpecOpts,
        ExecOptions(
            dry_run=spec.DebugSpecOpts.RunSpec.SpecExecOpts.dry_run,
            background=spec.DebugSpecOpts.RunSpec.SpecExecOpts.background,
            backend=spec.DebugSpecOpts.RunSpec.SpecExecOpts.backend,
            entrypoint=spec.DebugSpecOpts.RunSpec.SpecExecOpts.entrypoint,
            arguments=spec.DebugSpecOpts.RunSpec.SpecExecOpts.arguments,
        ),
    )

    code = DebugService().run(
        spec.DebugSpecOpts,
        ExecOptions(
            dry_run=spec.SpecExecOpts.dry_run,
            background=spec.SpecExecOpts.background,
            entrypoint=spec.SpecExecOpts.entrypoint,
            arguments=spec.SpecExecOpts.arguments,
            backend=spec.SpecExecOpts.backend,
            interactive=spec.SpecExecOpts.interactive,
        ),
    )

    if isinstance(handle, Popen):
        handle.communicate()

    return code


def debug_command(
    nds: Path | None,
    elf: Path | None,
    arguments: list[str] | None = None,
    backend: BackendOptions | None = None,
    background: bool = False,
    docker_screen: DockerOutputOptions | None = None,
    dry_run: bool = False,
    entrypoint: Path | None = None,
    port: int = 1024,
    settings: DebugSettings | None = None,
) -> int | Popen[bytes]:

    if settings is None:
        full_settings = load_settings()
        configure_logging_from_settings(settings=full_settings)
        settings = full_settings.debug

    log.debug("Settings for debug command:\n" f"- {str(settings)}\n")

    spec = create_debug_spec(
        nds=nds,
        elf=elf,
        arguments=arguments,
        backend=backend,
        background=background,
        docker_screen=docker_screen,
        dry_run=dry_run,
        entrypoint=entrypoint,
        port=port,
        settings=settings,
    )

    log.debug("Spec for debug command:\n" f"- {str(settings)}\n")

    return execute_debug(spec)



import subprocess

from bmde.core import logging
from bmde.core.docker import (
    can_run_docker,
    docker_remove_network,
    ensure_network_is_present,
)
from bmde.core.exec import ExecOptions, run_cmd
from bmde.core.types import DOCKER_DESMUME_DEBUG_NETWORK, DockerOutputOptions

from ..spec import RunSpecOpts
from .backend import RunBackend

log = logging.get_logger(__name__)


class DockerRunner(RunBackend):
    def is_available(self) -> bool:
        return can_run_docker()

    def run(
        self, spec: RunSpecOpts, exec_opts: ExecOptions
    ) -> int | subprocess.Popen[bytes]:
        docker_img = "aleixmt/desmume:latest"
        mounts = [
            "-v",
            f"{spec.nds_rom.parent}:/roms:ro",
            "-v",
            "desmume_docker_config:/home/desmume/.config/desmume",
        ]
        envs = []
        ports = []
        img_opt = []
        if spec.debug:
            ports += ["-p", "1024:1024"]  # Expose desmume to host on debug mode

        if spec.fat_image:
            mounts += ["-v", f"{spec.fat_image}:/fs/fat.img:rw"]
            img_opt += ["--cflash-image", "/fs/fat.img"]

        if spec.graphical_output == DockerOutputOptions.HOST:
            mounts += ["-v", "/tmp/.X11-unix:/tmp/.X11-unix"]
            envs += [
                "-e",
                "MODE=host",
                "-e",
                "DISPLAY=:0",
                "-e",
                "XVFB_DISPLAY=:99",
                "-e",
                "GEOMETRY=1024x768x24",
                "-e",
                "VNC_PORT=5900",
            ]
        if spec.graphical_output == DockerOutputOptions.VNC:
            ports += ["-p", "3000:3000", "-p", "3001:3001"]
            envs += ["-e", "MODE=vnc", "-e", "DISPLAY=:0"]

        entry = []
        if exec_opts.entrypoint:
            entry = ["--entrypoint", str(exec_opts.entrypoint)]

        debug_opt = []
        if spec.debug:
            if spec.arm9_debug_port is not None:
                debug_opt = [f"--arm9gdb={str(spec.arm9_debug_port)}"]
            else:
                debug_opt = ["--arm9gdb=1024"]

        arguments: list[str] = []
        if exec_opts.arguments is not None:
            arguments = list(exec_opts.arguments)

        if spec.nds_rom is not None:
            arguments += [f"/roms/{spec.nds_rom.name}"]

        network_opt: list[str] = []
        if spec.debug:
            network_opt += ["--name", "desmume", "--network", spec.docker_network]

        run_args = [
            "docker",
            "run",
            "--pull=always",
            "--rm",
        ]

        if exec_opts.interactive:
            run_args += ["-it"]

        run_args += [
            *network_opt,
            *mounts,
            *envs,
            *ports,
            *entry,
            docker_img,
            *img_opt,
            *debug_opt,
            *arguments,
        ]

        docker_net = None
        if spec.debug:
            docker_net = DOCKER_DESMUME_DEBUG_NETWORK
            if spec.docker_network is not None:
                docker_net = spec.docker_network

        if docker_net is not None:
            ensure_network_is_present(docker_net)

        handle = run_cmd(run_args, exec_opts)

        if docker_net is not None and not exec_opts.background:
            docker_remove_network(docker_net)

        return handle



from typing import Any

import typer

from bmde.commands.build.service import BuildService
from bmde.commands.debug.service import DebugService
from bmde.commands.git.service import GitService
from bmde.commands.patch.service import PatchService
from bmde.commands.run.service import RunService
from bmde.config.loader import load_settings
from bmde.config.schema import Settings
from bmde.core import logging
from bmde.core.logging import configure_logging_from_settings
from bmde.core.service import Service
from bmde.core.types import BackendOptions

log = logging.get_logger(__name__)


def check_command(settings: Settings) -> None:

    if settings is None:
        full_settings = load_settings()
        configure_logging_from_settings(
            settings=full_settings,
            obfuscate_sensibles=full_settings.logging.hide_sensibles,
        )

    services: list[tuple[str, Service[Any, Any], BackendOptions | None]] = [
        ("Git", GitService(), settings.git.execution_settings.backend),
        ("Run", RunService(), settings.run.execution_settings.backend),
        ("Build", BuildService(), settings.build.execution_settings.backend),
        ("Debug", DebugService(), settings.debug.execution_settings.backend),
        ("Patch", PatchService(), settings.patch.execution_settings.backend),
    ]

    msg = (
        "\nSummary of backend availability for the BMDE services in this machine (🚫 means not available, ✅ means "
        "available):\n"
    )
    for name, service, selected_backend in services:
        msg += f"\n  * Service: {name}\n"

        if selected_backend:
            backend_impl = service.mapping.get(selected_backend)
            if backend_impl:
                available = backend_impl.is_available()
                icon = "✅" if available else "🚫"
                msg += f"    Selected backend: {selected_backend.value} {icon}\n"
            else:
                msg += f"    Selected backend: {selected_backend.value} (Unknown backend?)\n"
        else:
            msg += "    Selected backend: Auto (None configured)\n"

        msg += "    Other backends for this service (by order of preference):\n"
        for backend_opt in service.order:
            if selected_backend and backend_opt == selected_backend:
                continue

            backend_impl = service.mapping.get(backend_opt)
            if backend_impl:
                available = backend_impl.is_available()
                icon = "✅" if available else "🚫"
                msg += f"      - {backend_opt.value}: {icon}\n"

    typer.echo(msg)



import subprocess

from bmde.core import logging

log = logging.get_logger(__name__)


def can_run_docker() -> bool:
    """
    Return True if the current user can run Docker containers, False otherwise.
    """
    # Try using the Docker CLI directly
    try:
        # Run 'docker info' quietly; suppress output
        subprocess.run(
            ["docker", "info"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True,
        )
        log.debug("Docker container runtime running")
        return True
    except FileNotFoundError:
        # Docker CLI not installed
        return False
    except subprocess.CalledProcessError:
        # Docker CLI exists but user cannot access the daemon
        return False
    except Exception:
        # Catch any unexpected errors
        return False


def docker_container_exists(container_name: str) -> bool:
    try:
        subprocess.check_output(
            ["docker", "inspect", container_name], stderr=subprocess.DEVNULL
        )
        return True
    except subprocess.CalledProcessError:
        return False


def docker_inspect_health(container_name: str) -> str | None:
    """
    Return health status string: "healthy", "unhealthy", "starting", or None if not found/no health.
    """
    try:
        out = subprocess.check_output(
            ["docker", "inspect", "-f", "{{.State.Health.Status}}", container_name],
            text=True,
            stderr=subprocess.STDOUT,
        ).strip()
        return out if out else None
    except subprocess.CalledProcessError:
        # Container absent or no Health configured
        return None


def docker_network_exists(network_name: str) -> bool:
    """
    Return True if a Docker network with the given name exists, False otherwise.
    """
    try:
        subprocess.check_output(
            ["docker", "network", "inspect", network_name], stderr=subprocess.DEVNULL
        )
        return True
    except subprocess.CalledProcessError:
        return False


def docker_create_network(network_name: str) -> None:
    """
    Create a Docker network with the given name if it does not already exist.
    """
    log.info(f"Creating Docker network '{network_name}'...")
    subprocess.run(["docker", "network", "create", network_name], check=True)


def docker_remove_network(network_name: str) -> None:
    """
    Remove a Docker network with the given name if it exists.
    """
    if docker_network_exists(network_name):
        log.info(f"Removing Docker network '{network_name}'...")
        subprocess.run(["docker", "network", "rm", network_name], check=True)


def ensure_network_is_present(network_name: str) -> None:
    if not docker_network_exists(network_name):
        docker_create_network(network_name)



import shutil
import subprocess

from bmde.core.exec import ExecOptions, run_cmd
from bmde.core.os_utils import is_command_available

from ..spec import PatchSpecOpts
from .backend import PatchBackend


class HostRunner(PatchBackend):
    def is_available(self) -> bool:
        return is_command_available("dlditool")

    def run(
        self, spec: PatchSpecOpts, exec_opts: ExecOptions
    ) -> int | subprocess.Popen[bytes]:
        if exec_opts.entrypoint is not None:
            entry = str(exec_opts.entrypoint)
        else:
            dlditool_path = shutil.which("dlditool")
            entry = "dlditool" if dlditool_path is None else dlditool_path
        args = [entry, str(spec.nds_rom)]
        if exec_opts.arguments is not None:
            args += list(exec_opts.arguments)
        return run_cmd(args, exec_opts)



"""
Defines the schema of settings of the application
"""

from pydantic import BaseModel

from bmde.commands.build.settings import BuildSettings
from bmde.commands.debug.settings import DebugSettings
from bmde.commands.git.settings import GitSettings
from bmde.commands.patch.settings import PatchSettings
from bmde.commands.run.settings import RunSettings
from bmde.config.logging_settings import LoggingSettings


class Settings(BaseModel):
    logging: LoggingSettings = LoggingSettings()
    run: RunSettings = RunSettings()
    build: BuildSettings = BuildSettings()
    git: GitSettings = GitSettings()
    patch: PatchSettings = PatchSettings()
    debug: DebugSettings = DebugSettings()



from dataclasses import dataclass
from pathlib import Path

from bmde.core.spec import BaseSpec
from bmde.core.spec_opts import SpecExecOpts


@dataclass
class GitSpecOpts(BaseSpec):
    d: Path
    ssh_username: str | None
    ssh_password: str | None
    ssh_host: str | None
    git_name: str | None
    git_email: str | None
    vpn_username: str | None
    vpn_password: str | None
    vpn_host: str | None
    vpn_port: int | None
    vpn_realm: str | None
    vpn_cert: str | None
    vpn_test_dns: str | None
    vpn_test_ip: str | None


@dataclass
class GitSpec(BaseSpec):
    SpecExecOpts: SpecExecOpts
    GitSpecOpts: GitSpecOpts



import rtoml
import typer

from bmde.config.loader import load_settings
from bmde.config.schema import Settings


def get_settings(default_only: bool) -> str:
    settings = Settings() if default_only else load_settings()

    # Convert to TOML and print
    toml_str = rtoml.dumps(settings.model_dump(mode="json", by_alias=True))

    # Post-process to comment out null values and add spacing
    lines: list[str] = []
    for i, line in enumerate(toml_str.splitlines()):
        stripped = line.strip()

        # Add blank line before section headers if missing
        if stripped.startswith("[") and i > 0 and lines and lines[-1].strip() != "":
            lines.append("")

        if stripped.endswith("= null") or stripped.endswith('= "null"'):
            lines.append(f"# {line}")
        else:
            lines.append(line)
    toml_str = "\n".join(lines)

    return toml_str


def show_config_callback(value: bool) -> None:
    if value:
        toml_str = get_settings(default_only=False)
        typer.echo(toml_str)
        raise typer.Exit()


def show_default_config_callback(value: bool) -> None:
    if value:
        toml_str = get_settings(default_only=True)
        final_str = "# This file can be generated with the shell command 'bmde --show-default-config > bmde.toml'\n\n"
        final_str += toml_str
        typer.echo(final_str)
        raise typer.Exit()



"""
Reusable CLI argument definition.
"""

from __future__ import annotations

from pathlib import Path
from typing import Annotated

import typer

from bmde.config.show_config import show_config_callback, show_default_config_callback
from bmde.core.types import BackendOptions, DockerOutputOptions

# Backend option for the rest of commands other than run
BackendOpt = Annotated[
    BackendOptions | None,
    typer.Option(
        "-b",
        "--backend",
        help="Backend to execute command:" " host|docker|flatpak",
        case_sensitive=False,
    ),
]

# Arguments passed directly to the entrypoint
ArgumentsOpt = Annotated[
    list[str] | None,
    typer.Argument(help="Arguments that are passed to the backend entrypoint."),
]

# Argument to send a directory
DirectoryOpt = Annotated[
    Path | None,
    typer.Option(
        "-d",
        "--directory",
        exists=True,
        file_okay=False,
        dir_okay=True,
        readable=True,
        resolve_path=True,
        help="Path to a directory. If omitted, the current directory is used.",
    ),
]


# Path to file that will be used as entrypoint
EntrypointOpt = Annotated[
    Path | None,
    typer.Option("--entrypoint", help="Override backend entrypoint executable"),
]

### Global options
# Dry-run flag
DryRunOpt = Annotated[
    bool,
    typer.Option("--dry-run", help="Simulate actions without executing", is_flag=True),
]

InteractiveOpt = Annotated[
    bool,
    typer.Option(
        "--interactive",
        help="Run in interactive mode (allocates TTY)",
    ),
]

LogFileOpt = Annotated[
    Path | None, typer.Option("-l", "--log-file", help="Path to log file (optional)")
]

ShowConfigOpt = Annotated[
    bool,
    typer.Option(
        "-s",
        "--show-config",
        callback=show_config_callback,
        is_eager=True,
        help="Show current configuration options. Can be used to freeze BMDE arguments into a bmde.toml configuration "
        "file",
        is_flag=True,
    ),
]

ShowDefaultConfigOpt = Annotated[
    bool,
    typer.Option(
        "-d",
        "--show-default-config",
        callback=show_default_config_callback,
        is_eager=True,
        help="Show default configuration options. Can be used as template of a bmde.toml configuration file",
        is_flag=True,
    ),
]

VerboseOpt = Annotated[
    bool,
    typer.Option(
        "-v", "--verbose", "--debug", help="Verbose output (debug)", is_flag=True
    ),
]

# Very verbose flag
VeryVerboseOpt = Annotated[
    bool,
    typer.Option("--t", "--trace", help="Very verbose output (trace)", is_flag=True),
]

# Quiet flag
QuietOpt = Annotated[
    bool,
    typer.Option("-q", "--quiet", help="Quiet mode (minimal output)", is_flag=True),
]

# Very quiet flag
VeryQuietOpt = Annotated[
    bool,
    typer.Option(
        "-Q", "--Quiet", "--no-output", help="Quiet mode (no output)", is_flag=True
    ),
]

ConfigOpt = Annotated[
    Path | None,
    typer.Option(
        "-c", "--config", help="Execution-specific config file (highest file priority)"
    ),
]

# --- Run options ---
# To specify debug in run options
DebugOpt = Annotated[
    bool, typer.Option("--debug", help="Enable GDB stub if supported", is_flag=True)
]

# To specify the debug port in options
PortOpt = Annotated[
    int, typer.Option("-p", "--port", help="Debug port (implies --debug)")
]

DockerNetworkOpt = Annotated[
    str | None,
    typer.Option(
        "-N",
        "--docker-network",
        "--network",
        help='Name of the Docker network to use when using the "docker" environment',
        is_flag=True,
    ),
]


DockerScreenOpt = Annotated[
    DockerOutputOptions | None,
    typer.Option(
        "-s",
        "--screen",
        help='Method to show the screen when using the "docker" environment',
        is_flag=True,
    ),
]


BackgroundOpt = Annotated[
    bool,
    typer.Option("-B", "--background", help="Run background", is_flag=True),
]

NdsRomOpt = Annotated[
    Path | None,
    typer.Option(
        "-n",
        "--nds",
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
        help="Path to the .nds binary (optional). If omitted, searches the current directory.",
    ),
]


ElfRomOpt = Annotated[
    Path | None,
    typer.Option(
        "-e",
        "--elf",
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
        help="Path to the .elf binary. If omitted, searches the current directory for an elf.",
    ),
]

FatImageOpt = Annotated[
    Path | None,
    typer.Option(
        "-i",
        "--image",
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
        help="Path to FAT image (optional)",
    ),
]

# Git


SshUsernameOpt = Annotated[
    str | None,
    typer.Option("--ssh-user", help="User name for the SSH authentication of git"),
]

SshPasswordOpt = Annotated[
    str | None,
    typer.Option(
        "--ssh-password", help="User password for the SSH authentication of git"
    ),
]

SshHostOpt = Annotated[
    str | None, typer.Option("--ssh-server", help="Hostname of the ssh server")
]

GitNameOpt = Annotated[
    str | None,
    typer.Option("--git-password", help="User name for git commit signature"),
]

GitEmailOpt = Annotated[
    str | None,
    typer.Option("--git-email", help="User email for git commit signature"),
]

VpnUsernameOpt = Annotated[
    str | None,
    typer.Option("--vpn-user", help="User name for forticlient authentication"),
]

VpnPasswordOpt = Annotated[
    str | None,
    typer.Option("--vpn-password", help="User password for forticlient authentication"),
]

VpnHostOpt = Annotated[
    str | None, typer.Option("--vpn-gateway", help="VPN gateway for forticlient")
]

VpnPortOpt = Annotated[
    int | None, typer.Option("--vpn-port", help="VPN port for forticlient")
]

VpnRealmOpt = Annotated[
    str | None, typer.Option("--vpn-realm", help="VPN realm for forticlient")
]

VpnCertOpt = Annotated[
    str | None, typer.Option("--vpn-cert", help="VPN cert for forticlient")
]

VpnTestDnsOpt = Annotated[
    str | None,
    typer.Option(
        "--vpn-test-dns",
        help="DNS direction that will be tested with an HTTP GET request to validate that we can access the "
        "internal "
        "services granted by the VPN and its implicit DNS resolution",
    ),
]

VpnTestIpOpt = Annotated[
    str | None,
    typer.Option(
        "--vpn-test-ip",
        help="IP direction that will be tested with an HTTP GET request to validate that we can access the internal "
        "services granted by the VPN",
    ),
]


# TODO Add all opts to this variable so they are marked as publicly exposed
__all__ = [
    "EntrypointOpt",
    "DebugOpt",
    "PortOpt",
    "DryRunOpt",
    "VerboseOpt",
    "QuietOpt",
    "DockerScreenOpt",
    "ConfigOpt",
    "BackendOpt",
    "ArgumentsOpt",
    "DirectoryOpt",
    "VeryVerboseOpt",
    "VeryQuietOpt",
    "LogFileOpt",
    "NdsRomOpt",
    "FatImageOpt",
    "SshUsernameOpt",
    "SshPasswordOpt",
    "SshHostOpt",
    "GitNameOpt",
    "GitEmailOpt",
    "VpnUsernameOpt",
    "VpnPasswordOpt",
    "VpnHostOpt",
    "VpnPortOpt",
    "VpnRealmOpt",
    "VpnCertOpt",
    "VpnTestDnsOpt",
    "VpnTestIpOpt",
    "ConfigOpt",
    "ElfRomOpt",
    "DockerNetworkOpt",
    "BackgroundOpt",
    "ShowConfigOpt",
    "InteractiveOpt",
    "ShowDefaultConfigOpt",
]



import shutil
import subprocess

from bmde.core import logging
from bmde.core.exec import ExecOptions, run_cmd
from bmde.core.os_utils import is_command_available

from ..spec import GitSpecOpts
from .backend import GitBackend

log = logging.get_logger(__name__)


class HostRunner(GitBackend):
    def is_available(self) -> bool:
        return is_command_available("git")

    def run(
        self, spec: GitSpecOpts, exec_opts: ExecOptions
    ) -> int | subprocess.Popen[bytes]:
        entry = (
            str(exec_opts.entrypoint) if exec_opts.entrypoint else (shutil.which("git"))
        )
        if not entry:
            return 127
        args = [entry]
        if exec_opts.arguments is not None:
            args += list(exec_opts.arguments)
        log.debug("Arguments for git in host backend: " + str(args))
        return run_cmd(args, exec_opts)



import shutil
import subprocess

from bmde.core import logging
from bmde.core.exec import ExecOptions, run_cmd
from bmde.core.os_utils import is_command_available

from ..spec import DebugSpecOpts
from .backend import DebugBackend

log = logging.get_logger(__name__)


class HostRunner(DebugBackend):
    def is_available(self) -> bool:
        return is_command_available("insight")

    def run(
        self, spec: DebugSpecOpts, exec_opts: ExecOptions
    ) -> int | subprocess.Popen[bytes]:
        if exec_opts.entrypoint is not None:
            entry = str(exec_opts.entrypoint)
        else:
            insight_path = shutil.which("desmume")
            entry = insight_path if insight_path is not None else "insight"
        args = [entry, str(spec.elf)]
        if exec_opts.arguments is not None:
            args += list(exec_opts.arguments)
        return run_cmd(args, exec_opts)



from __future__ import annotations

from bmde.core import logging
from bmde.core.service import Service

from ...core.types import BackendOptions
from .backends.backend import GitBackend
from .backends.docker import DockerRunner
from .backends.host import HostRunner
from .spec import GitSpecOpts

log = logging.get_logger(__name__)


class GitService(Service[GitSpecOpts, GitBackend]):
    def __init__(self) -> None:
        super().__init__(
            [BackendOptions.DOCKER, BackendOptions.HOST],
            {BackendOptions.HOST: HostRunner(), BackendOptions.DOCKER: DockerRunner()},
        )



from bmde.commands.build.service import BuildService
from bmde.core.types import BackendOptions


def test_build_service_initialization():
    """Verify that BuildService maps the correct backends."""
    service = BuildService()
    # Check if 'host' is in the mapping keys
    assert "host" in [key.value for key in service.mapping]


def test_choose_backend_force_env():
    """Verify that passing an env overrides the default order."""
    service = BuildService()
    # Assuming 'host' is a valid Enum value
    backends = service.choose_backend(BackendOptions.DOCKER)
    assert len(backends) == 1



import typer

from bmde.commands.patch.command import patch_command
from bmde.config.schema import Settings
from bmde.core import logging
from bmde.core.shared_options import (
    ArgumentsOpt,
    BackendOpt,
    BackgroundOpt,
    DirectoryOpt,
    DryRunOpt,
    EntrypointOpt,
    InteractiveOpt,
)

log = logging.get_logger(__name__)


def patch_controller(
    ctx: typer.Context,
    arguments: ArgumentsOpt = None,
    directory: DirectoryOpt = None,
    backend: BackendOpt = None,
    background: BackgroundOpt = False,
    interactive: InteractiveOpt = True,
    entrypoint: EntrypointOpt = None,
    dry_run: DryRunOpt = False,
) -> None:
    """dlditool wrapper. Patches a NDS ROM for FAT usage."""

    log.debug(
        "CLI options provided:\n"
        f"- Arguments: {str(arguments)}\n"
        f"- Directory: {str(directory)}\n"
        f"- Backend: {str(backend)}\n"
        f"- Background: {str(background)}\n"
        f"- Interactive: {str(interactive)}\n"
        f"- Entrypoint: {str(entrypoint)}\n"
        f"- Dry run: {str(dry_run)}\n"
    )

    settings: Settings = ctx.obj["settings"]

    ret = patch_command(
        d=directory,
        arguments=arguments,
        backend=backend,
        background=background,
        interactive=interactive,
        entrypoint=entrypoint,
        dry_run=dry_run,
        settings=settings.patch,
    )

    if isinstance(ret, int):
        raise typer.Exit(ret)



import os
import shutil
import subprocess

from bmde.core import logging
from bmde.core.exec import ExecOptions, run_cmd
from bmde.core.os_utils import is_command_available

from ..spec import BuildSpecOpts
from .backend import BuildBackend

log = logging.get_logger(__name__)


class HostRunner(BuildBackend):
    def is_available(self) -> bool:
        return self.check()

    def check(self) -> bool:
        checks_passed = True

        # Check make
        if not is_command_available("make"):
            log.error("'make' not found in PATH.")
            checks_passed = False
        else:
            log.debug("'make' found.")

        # Check devkitARM tools
        if not is_command_available("arm-none-eabi-gcc"):
            log.error("'arm-none-eabi-gcc' (devkitARM) not found in PATH.")
            checks_passed = False
        else:
            log.debug("'arm-none-eabi-gcc' found.")

        # Check DEVKITPRO env vartk

        if "DEVKITPRO" not in os.environ:
            log.warning(
                "DEVKITPRO environment variable not set. Some tools might fail."
            )

        # Check DEVKITARM env var
        if "DEVKITARM" not in os.environ:
            log.warning("DEVKITARM environment variable not set.")

        return checks_passed

    def run(
        self, spec: BuildSpecOpts, exec_opts: ExecOptions
    ) -> int | subprocess.Popen[bytes]:
        if exec_opts.entrypoint is not None:
            entry = str(exec_opts.entrypoint)
        else:
            make_path = shutil.which("make")
            entry = "make" if make_path is None else make_path
        args = [entry, str(spec.d)]
        if exec_opts.arguments is not None:
            args += list(exec_opts.arguments)
        return run_cmd(args, exec_opts)



from pydantic import BaseModel

from bmde.config.command_settings import ExecutionSettings


class VpnAuthSettings(BaseModel):
    enabled: bool = True
    username: str | None = None
    password: str | None = None
    host: str | None = None
    port: int | None = None
    cert: str | None = None
    realm: str | None = None
    test_dns: str | None = None
    test_ip: str | None = None


class GitSshSettings(BaseModel):
    username: str | None = None
    password: str | None = None
    host: str | None = None


class GitConfigSettings(BaseModel):
    name: str | None = None
    email: str | None = None


class GitSettings(BaseModel):
    execution_settings: ExecutionSettings = ExecutionSettings()

    git: GitConfigSettings = GitConfigSettings()
    ssh: GitSshSettings = GitSshSettings()
    vpn: VpnAuthSettings = VpnAuthSettings()



import subprocess

from bmde.core import logging
from bmde.core.docker import (
    can_run_docker,
    docker_remove_network,
    ensure_network_is_present,
)
from bmde.core.exec import ExecOptions, run_cmd
from bmde.core.types import DOCKER_DESMUME_DEBUG_NETWORK, DockerOutputOptions

from ..spec import DebugSpecOpts
from .backend import DebugBackend

log = logging.get_logger(__name__)


class DockerRunner(DebugBackend):
    def is_available(self) -> bool:
        return can_run_docker()

    def run(
        self, spec: DebugSpecOpts, exec_opts: ExecOptions
    ) -> int | subprocess.Popen[bytes]:

        docker_img = "aleixmt/insight:latest"
        mounts = [
            "-v",
            f"{spec.elf.parent}:/input:ro",
        ]
        envs = []
        ports = []

        if spec.docker_screen == DockerOutputOptions.HOST:
            mounts += ["-v", "/tmp/.X11-unix:/tmp/.X11-unix"]
            envs += [
                "-e",
                "MODE=host",
                "-e",
                "DISPLAY=:0",
                "-e",
                "XVFB_DISPLAY=:99",
                "-e",
                "GEOMETRY=1024x768x24",
                "-e",
                "VNC_PORT=5900",
            ]
        if spec.docker_screen == DockerOutputOptions.VNC:
            ports += ["-p", "3000:3000", "-p", "3001:3001"]
            envs += ["-e", "MODE=vnc", "-e", "DISPLAY=:0"]
        entry = []
        if exec_opts.entrypoint:
            entry = ["--entrypoint", str(exec_opts.entrypoint)]

        arguments: list[str] = []
        if spec.elf is not None:
            arguments += [f"/input/{spec.elf.name}"]

        if exec_opts.arguments is not None:
            arguments += list(exec_opts.arguments)

        docker_net = DOCKER_DESMUME_DEBUG_NETWORK
        if spec.docker_network is not None:
            docker_net = spec.docker_network
        network = ["--network", docker_net]

        run_args = [
            "docker",
            "run",
            "--pull=always",
            "--rm",
            "-it",
            "--cap-add=SYS_PTRACE",
            "--security-opt",
            "seccomp=unconfined",
            *network,
            *mounts,
            *envs,
            *ports,
            *entry,
            docker_img,
            *arguments,
        ]

        ensure_network_is_present(docker_net)

        handle = run_cmd(run_args, exec_opts)

        if not exec_opts.background:
            docker_remove_network(docker_net)

        return handle



from dataclasses import dataclass
from pathlib import Path

from bmde.core.spec import BaseSpec
from bmde.core.spec_opts import SpecExecOpts


@dataclass
class PatchSpecOpts(BaseSpec):
    nds_rom: Path


@dataclass
class PatchSpec(BaseSpec):
    SpecExecOpts: SpecExecOpts
    PatchSpecOpts: PatchSpecOpts



import typer

from bmde.commands.run.command import run_command
from bmde.config.schema import Settings
from bmde.core import logging
from bmde.core.shared_options import (
    ArgumentsOpt,
    BackendOpt,
    BackgroundOpt,
    DebugOpt,
    DirectoryOpt,
    DockerNetworkOpt,
    DockerScreenOpt,
    DryRunOpt,
    EntrypointOpt,
    FatImageOpt,
    InteractiveOpt,
    NdsRomOpt,
    PortOpt,
)

log = logging.get_logger(__name__)


def run_controller(
    ctx: typer.Context,
    nds_rom: NdsRomOpt = None,
    directory: DirectoryOpt = None,
    arguments: ArgumentsOpt = None,
    arm9_debug_port: PortOpt = 1024,
    backend: BackendOpt = None,
    background: BackgroundOpt = False,
    interactive: InteractiveOpt = True,
    debug: DebugOpt = False,
    docker_network: DockerNetworkOpt = None,
    dry_run: DryRunOpt = False,
    entrypoint: EntrypointOpt = None,
    fat_image: FatImageOpt = None,
    graphical_output: DockerScreenOpt = None,
) -> None:

    log.debug(
        "CLI options provided:\n"
        "- Arguments:\n"
        f"- NDS ROM: {str(nds_rom)}\n"
        f"- Rom directory: {str(directory)}\n"
        "- Behavioural parameters:\n"
        f"- Arguments: {str(arguments)}\n"
        f"- ARM 9 debug port: {str(arm9_debug_port)}\n"
        f"- Backend: {str(backend)}\n"
        f"- Background: {str(background)}\n"
        f"- Interactive: {str(interactive)}\n"
        f"- Debug: {str(debug)}\n"
        f"- Docker Network: {str(docker_network)}\n"
        f"- Dry run: {str(dry_run)}\n"
        f"- Entrypoint: {str(entrypoint)}\n"
        f"- FAT image: {str(fat_image)}\n"
        f"- Graphical output: {str(graphical_output)}\n"
    )

    settings: Settings = ctx.obj["settings"]

    log.debug("Loaded settings:\n" f"{str(settings.run)}\n")
    run_command(
        nds_rom=nds_rom,
        directory=directory,
        arguments=arguments,
        arm9_debug_port=arm9_debug_port,
        backend=backend,
        background=background,
        interactive=interactive,
        debug=debug,
        docker_network=docker_network,
        dry_run=dry_run,
        entrypoint=entrypoint,
        fat_image=fat_image,
        graphical_output=graphical_output,
        settings=settings.run,
    )



from pathlib import Path

from pydantic import BaseModel

from bmde.core.types import LogLevel, get_default_log_path


class LoggingSettings(BaseModel):
    level: LogLevel = LogLevel.get_default_log_level()
    file: Path | None = get_default_log_path()
    hide_sensibles: bool | None = True



from dataclasses import dataclass
from pathlib import Path

from bmde.core.spec import BaseSpec
from bmde.core.spec_opts import SpecExecOpts


@dataclass
class BuildSpecOpts(BaseSpec):
    d: Path


@dataclass
class BuildSpec(BaseSpec):
    SpecExecOpts: SpecExecOpts
    BuildSpecOpts: BuildSpecOpts





from pathlib import Path

from bmde.core import logging

log = logging.get_logger(__name__)


def validate_file(p: Path) -> Path:
    if not p.is_file():
        raise FileNotFoundError(f"NDS file not found: {p}")
    if p.suffix.lower() != ".nds":
        raise ValueError(f"Expected a .nds file, got: {p.name}")
    return p


def discover_files_in_dir(search_dir: Path, glob: str) -> tuple[Path, bool]:
    files = sorted(search_dir.glob(glob))
    if not files:
        raise FileNotFoundError(f"No .nds file found in {search_dir}")
    if len(files) > 1:
        log.warning(
            "Multiple .nds files found in %s; assuming '%s'.",
            search_dir,
            files[0].name,
        )
    log.info(f"Found {len(files)} .nds files in {search_dir}")
    return files[0], True


def resolve_nds(maybe_nds: Path | None, cwd: Path) -> tuple[Path, bool]:
    """
    Return (nds_path, assumed_flag). If maybe_nds is None, discover in cwd.
    """
    if maybe_nds is not None:
        return validate_file(maybe_nds), False
    return discover_files_in_dir(cwd, "*.nds")


def resolve_elf(maybe_elf: Path | None, cwd: Path) -> tuple[Path, bool]:
    """
    Return (nds_path, assumed_flag). If maybe_nds is None, discover in cwd.
    """
    if maybe_elf is not None:
        return validate_file(maybe_elf), False
    return discover_files_in_dir(cwd, "*.elf")



import logging
import shutil
import subprocess
from pathlib import Path

import pytest

from bmde.commands.build.command import build_command
from bmde.core.logging import setup_logging


def test_build_smoke(tmp_path: Path):
    """
    Smoke test for the build command.
    Clones a hello-world repo and attempts to build it.
    This is an integration test.
    """
    # Setup logging to see output
    setup_logging(level=logging.DEBUG)

    # Check prerequisites
    if not shutil.which("git"):
        pytest.skip("git not found")

    repo_url = "https://github.com/URV-teacher/hello-world-nds"
    repo_dir = tmp_path / "hello-world-nds"

    print(f"Cloning {repo_url} into {repo_dir}...")
    # Clone repo
    try:
        # Removed capture_output=True to let git output be seen (with pytest -s)
        subprocess.run(["git", "clone", repo_url, str(repo_dir)], check=True)
    except subprocess.CalledProcessError as e:
        pytest.fail(f"Failed to clone repo: {e}")

    # Run build command
    # We call the API directly to avoid CLI overhead
    print(f"Building in {repo_dir}")

    ret = build_command(d=repo_dir, interactive=False)

    ret_code = 0
    if isinstance(ret, int):
        ret_code = ret
    else:
        # It's a Popen object
        ret.communicate()
        ret_code = ret.returncode

    assert ret_code == 0, "Build failed with non-zero exit code"



from __future__ import annotations

from bmde.core import logging
from bmde.core.service import Service

from ...core.types import BackendOptions
from .backends.backend import PatchBackend
from .backends.docker import DockerRunner
from .backends.host import HostRunner
from .spec import PatchSpecOpts

log = logging.get_logger(__name__)


class PatchService(Service[PatchSpecOpts, PatchBackend]):
    def __init__(self) -> None:
        super().__init__(
            [BackendOptions.DOCKER, BackendOptions.HOST],
            {BackendOptions.HOST: HostRunner(), BackendOptions.DOCKER: DockerRunner()},
        )



import logging
import os
import shutil
import subprocess
import time
import urllib.request
import zipfile
from pathlib import Path

import pytest

from bmde.commands.build.command import build_command
from bmde.commands.patch.command import patch_command
from bmde.commands.run.command import run_command
from bmde.core.logging import setup_logging
from bmde.core.types import DockerOutputOptions


def test_run_filesystem(tmp_path: Path):
    """
    Integration test for the run command with filesystem interaction.
    1. Clones filesystem-nds repo.
    2. Builds it.
    3. Patches the ROM.
    4. Downloads a FAT image.
    5. Runs the ROM with the FAT image.
    6. Verifies output.txt is written to the FAT image.
    """
    setup_logging(level=logging.DEBUG)

    # Prerequisites
    if not shutil.which("git"):
        pytest.skip("git not found")

    # 1. Clone
    repo_url = "https://github.com/URV-teacher/filesystem-nds"
    repo_dir = tmp_path / "filesystem-nds"
    print(f"Cloning {repo_url}...")
    subprocess.run(["git", "clone", repo_url, str(repo_dir)], check=True)

    # 2. Build
    print("Building...")
    ret = build_command(d=repo_dir, interactive=False)
    if isinstance(ret, subprocess.Popen):
        ret.wait()
        assert ret.returncode == 0, "Build failed"
    elif isinstance(ret, int):
        assert ret == 0, "Build failed"

    # Find .nds file
    nds_files = list(repo_dir.glob("*.nds"))
    assert len(nds_files) > 0, "No .nds file found after build"
    nds_file = nds_files[0]
    print(f"Found ROM: {nds_file}")

    # 3. Patch
    print("Patching...")
    ret = patch_command(d=repo_dir, interactive=False)
    if isinstance(ret, subprocess.Popen):
        ret.wait()
        assert ret.returncode == 0, "Patch failed"
    elif isinstance(ret, int):
        assert ret == 0, "Patch failed"

    # 4. Prepare FAT image
    print("Downloading FAT image...")
    fat_zip_url = (
        "https://raw.githubusercontent.com/URV-teacher/desmume-docker/master/fs/fat.zip"
    )
    fat_zip_path = tmp_path / "fat.zip"
    try:
        urllib.request.urlretrieve(fat_zip_url, fat_zip_path)
    except Exception as e:
        pytest.fail(f"Failed to download FAT image: {e}")

    print("Extracting FAT image...")
    with zipfile.ZipFile(fat_zip_path, "r") as zip_ref:
        zip_ref.extractall(tmp_path)

    # Locate extracted .img file
    fat_imgs = list(tmp_path.glob("*.img"))
    if not fat_imgs:
        # Try looking inside extracted folders if any
        fat_imgs = list(tmp_path.rglob("*.img"))

    assert len(fat_imgs) > 0, "No .img file found in extracted zip"
    fat_img_path = fat_imgs[0]
    os.chmod(fat_img_path, 0o777)
    print(f"Using FAT image: {fat_img_path}")

    # 5. Run
    print("Running emulator...")
    # Run in background so we can stop it
    proc = run_command(
        nds_rom=nds_file,
        fat_image=fat_img_path,
        interactive=False,
        graphical_output=DockerOutputOptions.VNC,
    )

    # Wait for emulation to start and write to file
    # The ROM is simple, it should write quickly.
    time.sleep(10)

    # Terminate emulator
    if isinstance(proc, subprocess.Popen):
        print("Terminating emulator...")
        proc.terminate()
        try:
            proc.wait(timeout=5)
        except subprocess.TimeoutExpired:
            proc.kill()

    # 6. Verify content
    print("Verifying output...")
    expected_text = (
        "Hola, este es un test en DeSmuME.\nProbando escritura en archivo.\n"
    )

    # Method A: mtools (preferred if available)
    if shutil.which("mtype"):
        print("Using mtools to verify...")
        try:
            # mtype -i image ::file
            result = subprocess.run(
                ["mtype", "-i", str(fat_img_path), "::output.txt"],
                capture_output=True,
                text=True,
            )
            if result.returncode == 0:
                content = result.stdout
                # Normalize newlines just in case
                if expected_text in content:
                    print("Verification successful (mtools)!")
                    return
                else:
                    print(f"Content mismatch via mtools. Got: {content!r}")
            else:
                print(f"mtype failed: {result.stderr}")
        except Exception as e:
            print(f"mtools check failed: {e}")

    # Method B: Binary search (Fallback)
    print("Using binary search to verify...")
    with open(fat_img_path, "rb") as f:
        data = f.read()

    # The text in the file might use \n or \r\n. The expected string has \n.
    # We search for the main sentence.
    search_phrase = b"Hola, este es un test en DeSmuME"

    if search_phrase in data:
        print("Verification successful (binary search)!")
    else:
        pytest.fail(f"Expected text '{search_phrase.decode()}' not found in FAT image.")



"""
The responsibility of this file and the similar ones under each command, is to translate the logic of the bmde interface
into the specification of the pure logic of the command, for example, translating shell into a corresponding entrypoint
"""

from __future__ import annotations

from pathlib import Path
from subprocess import Popen

from bmde.core import logging
from bmde.core.exec import ExecOptions

from ...config.loader import load_settings
from ...core.file_utils import resolve_nds
from ...core.logging import configure_logging_from_settings
from ...core.spec_opts import SpecExecOpts
from ...core.types import (
    DOCKER_DESMUME_DEBUG_NETWORK,
    BackendOptions,
    DockerOutputOptions,
)
from .service import RunService
from .settings import RunSettings
from .spec import RunSpec, RunSpecOpts

log = logging.get_logger(__name__)


def create_run_spec(
    nds_rom: Path | None,
    directory: Path | None,
    arguments: list[str] | None = None,
    arm9_debug_port: int | None = 1000,
    backend: BackendOptions | None = None,
    background: bool | None = False,
    interactive: bool = True,
    debug: bool | None = False,
    dry_run: bool = False,
    entrypoint: Path | None = None,
    docker_network: str | None = None,
    fat_image: Path | None = None,
    graphical_output: DockerOutputOptions | None = None,
    settings: RunSettings | None = None,
) -> RunSpec:
    """
    Creates the RunSpec from the provided arguments.
    Useful for the API to generate a spec without executing it immediately.
    """

    if directory is None:
        directory = Path.cwd()

    nds_resolved, _ = resolve_nds(nds_rom, cwd=directory)

    if settings is None:
        settings = RunSettings()

    # Default args are prepended to CLI args, not overridden
    passed_args: list[str] = []
    if arguments is not None:
        passed_args += arguments
    if settings.execution_settings.arguments is not None:
        passed_args += settings.execution_settings.arguments

    return RunSpec(
        RunSpecOpts=RunSpecOpts(
            nds_rom=nds_resolved,
            arm9_debug_port=(
                arm9_debug_port
                if arm9_debug_port is not None
                else settings.arm9_debug_port
            ),
            debug=(
                debug
                if debug is not None
                else (settings.debug if settings.debug is not None else False)
            ),
            docker_network=(
                docker_network
                if docker_network is not None
                else (
                    settings.docker_network
                    if settings.docker_network is not None
                    else DOCKER_DESMUME_DEBUG_NETWORK
                )
            ),
            fat_image=fat_image if fat_image is not None else settings.fat_image,
            graphical_output=(
                graphical_output
                if graphical_output is not None
                else settings.graphical_output
            ),
        ),
        SpecExecOpts=SpecExecOpts(
            backend=(
                backend
                if backend is not None
                else (
                    settings.execution_settings.backend
                    if settings.execution_settings.backend is not None
                    else BackendOptions.DOCKER
                )
            ),
            background=(
                background
                if background is not None
                else (
                    settings.execution_settings.background
                    if settings.execution_settings.background is not None
                    else False
                )
            ),
            interactive=(
                interactive
                if interactive is not None
                else settings.execution_settings.interactive
            ),
            dry_run=(
                dry_run if dry_run is not None else settings.execution_settings.dry_run
            ),
            entrypoint=(
                entrypoint
                if entrypoint is not None
                else settings.execution_settings.entrypoint
            ),
            arguments=passed_args,
        ),
    )


def execute_run(
    spec: RunSpec,
) -> int | Popen[bytes]:
    """
    Executes the RunSpec.
    The API can call this directly with a pre-constructed spec.
    """
    handle = RunService().run(
        spec.RunSpecOpts,
        ExecOptions(
            backend=spec.SpecExecOpts.backend,
            dry_run=spec.SpecExecOpts.dry_run,
            background=spec.SpecExecOpts.background,
            interactive=spec.SpecExecOpts.interactive,
            entrypoint=spec.SpecExecOpts.entrypoint,
            arguments=spec.SpecExecOpts.arguments,
        ),
    )

    return handle


def run_command(
    nds_rom: Path | None = None,
    directory: Path | None = None,
    arguments: list[str] | None = None,
    arm9_debug_port: int | None = 1024,
    backend: BackendOptions | None = None,
    background: bool | None = False,
    interactive: bool = True,
    debug: bool = False,
    docker_network: str | None = None,
    dry_run: bool = False,
    entrypoint: Path | None = None,
    fat_image: Path | None = None,
    graphical_output: DockerOutputOptions | None = None,
    settings: RunSettings | None = None,
) -> int | Popen[bytes]:
    """
    CLI Entrypoint. Orchestrates spec creation and execution.
    """

    if settings is None:
        full_settings = load_settings()
        configure_logging_from_settings(settings=full_settings)
        settings = full_settings.run

    spec = create_run_spec(
        nds_rom=nds_rom,
        directory=directory,
        arguments=arguments,
        arm9_debug_port=arm9_debug_port,
        backend=backend,
        background=background,
        interactive=interactive,
        debug=debug,
        docker_network=docker_network,
        dry_run=dry_run,
        entrypoint=entrypoint,
        fat_image=fat_image,
        graphical_output=graphical_output,
        settings=settings,
    )

    return execute_run(spec)



import os
import shutil
import subprocess


def is_command_available(cmd: str) -> bool:
    """
    Return True if the given command is available in the system PATH.

    Args:
        cmd (str): The command to check, e.g. 'docker', 'make', 'git'.

    Returns:
        bool: True if the command is found, False otherwise.
    """
    return shutil.which(cmd) is not None


def host_uid_gid() -> tuple[int, int]:
    """Return (uid, gid) if available on this OS; otherwise None."""
    # Prefer Python stdlib where available (POSIX)
    if hasattr(os, "getuid") and hasattr(os, "getgid"):
        try:
            return os.getuid(), os.getgid()
        except Exception:
            return 1000, 1000
    # Fallback to `id` command (e.g., inside POSIX shells without getuid support).
    try:
        uid = subprocess.check_output(["id", "-u"], text=True).strip()
        gid = subprocess.check_output(["id", "-g"], text=True).strip()
        return int(uid), int(gid)
    except Exception:
        return 1000, 1000



import os
import subprocess
import time
from pathlib import Path

from bmde.core import logging
from bmde.core.docker import (
    can_run_docker,
    docker_container_exists,
    docker_inspect_health,
)
from bmde.core.exec import ExecOptions, run_cmd
from bmde.core.os_utils import host_uid_gid

from ..spec import GitSpecOpts
from .backend import GitBackend

log = logging.get_logger(__name__)


def _run_vpn(
    spec: GitSpecOpts, exec_opts: ExecOptions, container_name: str
) -> int | None | subprocess.Popen[bytes]:
    docker_img = "aleixmt/forticlient:latest"

    envs: list[str] = [
        "-e",
        f"SSH_USERNAME={spec.ssh_username}",
        "-e",
        f"SSH_PASSWORD={spec.ssh_password}",
        "-e",
        f"SSH_HOST={spec.ssh_host}",
        "-e",
        f"GIT_NAME={spec.git_name}",
        "-e",
        f"GIT_EMAIL={spec.git_email}",
        "-e",
        f"VPN_USERNAME={spec.vpn_username}",
        "-e",
        f"VPN_PASSWORD={spec.vpn_password}",
        "-e",
        f"VPN_HOST={spec.vpn_host}",
        "-e",
        f"VPN_PORT={spec.vpn_port}",
        "-e",
        f"VPN_REALM={spec.vpn_realm}",
        "-e",
        f"VPN_CERT={spec.vpn_cert}",
        "-e",
        f"VPN_TEST_DNS={spec.vpn_test_dns}",
        "-e",
        f"VPN_TEST_IP={spec.vpn_test_ip}",
    ]

    run_args = [
        "docker",
        "run",
        "--pull=always",
        "-d",
        "--rm",
    ]

    if exec_opts.interactive:
        run_args += ["-it"]

    run_args += [
        "--name",
        container_name,
        "--cap-add",
        "NET_ADMIN",
        "--device",
        "/dev/ppp",
        "--health-cmd=[ -f /READY ] || exit 1",
        "--health-interval=5s",
        "--health-timeout=2s",
        "--health-retries=12",
        "--health-start-period=120s",
        *envs,
        docker_img,
    ]

    return run_cmd(run_args, exec_opts)


def _ensure_vpn_healthy(
    spec: GitSpecOpts, exec_opts: ExecOptions, timeout_s: int = 20000
) -> None:
    """
    Ensure the 'forti-vpn' service is running and HEALTHY.
    If absent or not healthy, start with `docker compose up -d forti-vpn` and wait.
    """
    container_name = "forti-vpn"
    status = docker_inspect_health(container_name)

    log.debug("VPN container status: " + str(status))
    if status == "healthy":
        log.debug("VPN is healthy")
        return
    elif docker_container_exists(container_name):
        log.debug(
            f"Container '{container_name}' exists but is not healthy ({status}); removing it..."
        )
        try:
            subprocess.run(
                ["docker", "rm", "-f", container_name], check=True, text=True
            )
        except subprocess.CalledProcessError as e:
            raise RuntimeError(
                f"Failed to remove unhealthy container '{container_name}': {e}"
            ) from e

    log.info("Running VPN")
    _run_vpn(spec, exec_opts, container_name)

    # Poll until healthy or timeout
    previous_state = None  # Will be used to determine if the container started and stopped, which means failure
    deadline = time.time() + timeout_s
    while time.time() < deadline:
        status = docker_inspect_health(container_name)
        if status == "healthy":
            log.info("VPN is healthy.")
            return
        elif status == "starting":
            log.info(
                f"VPN health: {status}. Please, wait. The VPN connection usually takes 120 seconds."
            )
            previous_state = status
        else:
            log.info(
                f"VPN health: {status}. Please, wait. The VPN connection usually takes 120 seconds."
            )
            if previous_state == "starting":
                log.error(
                    f"'{container_name}' was starting and now it is not healthy, meaning the container could not "
                    f"start. Your configuration may be wrong. Aborting..."
                )
                exit(1)
            previous_state = status
        time.sleep(10)

    raise TimeoutError(
        f"'{container_name}' did not become healthy within {timeout_s} seconds"
    )


# --------------------------- Git Runner ----------------------------


class DockerRunner(GitBackend):
    def is_available(self) -> bool:
        return can_run_docker()

    def run(
        self, spec: GitSpecOpts, exec_opts: ExecOptions
    ) -> int | subprocess.Popen[bytes]:
        """
        1) Ensure 'forti-vpn' compose service is up & healthy (start in background if needed).
        2) Run the git container sharing the network namespace with 'forti-vpn' with the command.
        """

        docker_img = "aleixmt/git-sshpass:latest"

        # 2) Ensure VPN is healthy (starts `docker compose up -d forti-vpn` if needed)
        _ensure_vpn_healthy(
            spec, exec_opts, timeout_s=getattr(spec, "vpn_timeout_s", 300)
        )

        # 3) Build docker run args for the git container
        entry = []
        if exec_opts.entrypoint:
            entry = ["--entrypoint", str(exec_opts.entrypoint)]

        # primary project mount (writable)
        host_path = str(Path(spec.d).resolve())
        dirname = os.path.basename(host_path)

        # TODO mount logs
        mounts = ["-v", f"{spec.d}:/repos/{dirname}:rw"]
        workdir_opt = ["-w", f"/repos/{dirname}"]

        # share network namespace with the running VPN container
        net = ["--network", "container:forti-vpn"]
        uid, gid = host_uid_gid()
        user_opt = ["--user", f"{uid}:{gid}"]

        envs: list[str] = [
            "-e",
            f"SSH_USERNAME={spec.ssh_username}",
            "-e",
            f"SSH_PASSWORD={spec.ssh_password}",
            "-e",
            f"SSH_HOST={spec.ssh_host}",
            "-e",
            f"GIT_NAME={spec.git_name}",
            "-e",
            f"GIT_EMAIL={spec.git_email}",
        ]

        args: list[str] = []
        if exec_opts.arguments:
            args = exec_opts.arguments

        run_args = [
            "docker",
            "run",
            "--pull=always",
            "--rm",
        ]

        if exec_opts.interactive:
            run_args += ["-it"]

        run_args += [
            *user_opt,
            *net,
            *mounts,
            *envs,
            *entry,
            *workdir_opt,
            docker_img,
            *args,
        ]

        log.debug("Run args for Docker" + str(run_args))
        return run_cmd(run_args, exec_opts)



from __future__ import annotations

from pathlib import Path
from subprocess import Popen

from bmde.core import logging
from bmde.core.exec import ExecOptions

from ...config.loader import load_settings
from ...core.file_utils import resolve_nds
from ...core.logging import configure_logging_from_settings
from ...core.spec_opts import SpecExecOpts
from ...core.types import BackendOptions
from .service import PatchService
from .settings import PatchSettings
from .spec import PatchSpec, PatchSpecOpts

log = logging.get_logger(__name__)


def create_patch_spec(
    d: Path | None,
    nds_rom: Path | None = None,
    arguments: list[str] | None = None,
    backend: BackendOptions | None = None,
    background: bool = False,
    dry_run: bool = False,
    entrypoint: Path | None = None,
    interactive: bool = True,
    settings: PatchSettings | None = None,
) -> PatchSpec:

    if d is None:
        d = Path.cwd()

    nds_resolved, _ = resolve_nds(nds_rom, cwd=d)

    if settings is None:
        settings = PatchSettings()

    return PatchSpec(
        PatchSpecOpts=PatchSpecOpts(nds_rom=nds_resolved),
        SpecExecOpts=SpecExecOpts(
            backend=(
                backend
                if backend is not None
                else (
                    settings.execution_settings.backend
                    if settings.execution_settings.backend is not None
                    else BackendOptions.DOCKER
                )
            ),
            entrypoint=(
                entrypoint
                if entrypoint is not None
                else settings.execution_settings.entrypoint
            ),
            arguments=arguments,
            dry_run=(
                dry_run if dry_run is not None else settings.execution_settings.dry_run
            ),
            background=(
                background
                if background is not None
                else settings.execution_settings.background
            ),
            interactive=interactive,
        ),
    )


def execute_patch(spec: PatchSpec) -> int | Popen[bytes]:
    return PatchService().run(
        spec.PatchSpecOpts,
        ExecOptions(
            dry_run=spec.SpecExecOpts.dry_run,
            background=spec.SpecExecOpts.background,
            entrypoint=spec.SpecExecOpts.entrypoint,
            arguments=spec.SpecExecOpts.arguments,
            backend=spec.SpecExecOpts.backend,
            interactive=spec.SpecExecOpts.interactive,
        ),
    )


def patch_command(
    d: Path | None,
    nds_rom: Path | None = None,
    arguments: list[str] | None = None,
    backend: BackendOptions | None = None,
    background: bool = False,
    dry_run: bool = False,
    entrypoint: Path | None = None,
    interactive: bool = True,
    settings: PatchSettings | None = None,
) -> int | Popen[bytes]:

    if settings is None:
        full_settings = load_settings()
        configure_logging_from_settings(settings=full_settings)
        settings = full_settings.patch

    spec = create_patch_spec(
        d=d,
        nds_rom=nds_rom,
        arguments=arguments,
        backend=backend,
        background=background,
        dry_run=dry_run,
        entrypoint=entrypoint,
        interactive=interactive,
        settings=settings,
    )

    return execute_patch(spec)



from __future__ import annotations

from bmde.core import logging
from bmde.core.service import Service
from bmde.core.types import BackendOptions as RunBackendName

from .backends.backend import DebugBackend
from .backends.docker import DockerRunner
from .backends.host import HostRunner
from .spec import DebugSpecOpts

log = logging.get_logger(__name__)


class DebugService(Service[DebugSpecOpts, DebugBackend]):
    def __init__(self) -> None:
        super().__init__(
            [RunBackendName.HOST, RunBackendName.DOCKER],
            {
                RunBackendName.DOCKER: DockerRunner(),
                RunBackendName.HOST: HostRunner(),
            },
        )
